# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os

import pyworkflow.viewer as pwviewer
import pyworkflow.utils as pwutils

from pwem import Domain
from pwem import emlib
import pwem.objects as emobj
import pwem.protocols as emprot

from .views import (ObjectView, MicrographsView, CoordinatesObjectView,
                    ClassesView, Classes3DView, CtfView, DataView)
from .showj import (RENDER, SAMPLINGRATE, ORDER, VISIBLE, MODE, MODE_MD,
                    SORT_BY, getJvmMaxMemory, launchTiltPairPickerGUI, LABELS)
from ..convert.headers import Ccp4Header


class RegistryViewerConfig:
    config = {}

    @classmethod
    def getConfig(cls, type):
        return cls.config.get(type, None)

    @classmethod
    def registerConfig(cls, type, config=None):
        """
        Allow registration of other objects
        :param type: Class to register --> SetOfXXX
        :param config: Optional, dictionary with the fields configuration, otherwise all fields will be shown.

        :return: Nothing

        """
        if config is None:
            config = {MODE: MODE_MD}

        cls.config[type] = config

# Registering viewer config
RegistryViewerConfig.registerConfig(emobj.SetOfPDBs, {ORDER: 'id _filename ',
                                   VISIBLE: 'id _filename ',
                                   MODE: MODE_MD,
                                   RENDER: "no"})


#SetOfParticles
labels = ('id enabled _index _filename _xmipp_zScore _xmipp_cumulativeSSNR '
                      '_sampling _xmipp_scoreByVariance _xmipp_scoreEmptiness '
                      '_ctfModel._defocusU _ctfModel._defocusV _ctfModel._defocusAngle '
                      '_transform._matrix')
RegistryViewerConfig.registerConfig(emobj.SetOfParticles,
                                   {ORDER: labels,
                                   VISIBLE: labels,
                                    SORT_BY: '_xmipp_zScore asc',
                                    RENDER: '_filename'})


# SetOfVolumes
labels = 'id enabled comment _filename '
RegistryViewerConfig.registerConfig(emobj.SetOfVolumes,
                                   {ORDER: labels,
                                   VISIBLE: labels,
                                    SORT_BY: '_xmipp_zScore asc',
                                    RENDER: '_filename'})

# SetOfMovies
labels = 'id _filename _samplingRate _acquisition._dosePerFrame _acquisition._doseInitial '
RegistryViewerConfig.registerConfig(emobj.SetOfMovies,
                                   {ORDER: labels,
                                   VISIBLE: labels,
                                    MODE: MODE_MD,
                                    RENDER: 'no'})

#MicrographsTiltPair
labels = 'id enabled _untilted._filename _tilted._filename'
renderLabels = '_untilted._filename _tilted._filename'

RegistryViewerConfig.registerConfig(emobj.MicrographsTiltPair,
                                   {ORDER: labels,
                                   VISIBLE: labels,
                                    MODE: MODE_MD,
                                    RENDER: renderLabels})


#ParticlesTiltPair
labels = 'id enabled _untilted._filename _tilted._filename'
renderLabels = '_untilted._filename _tilted._filename'
RegistryViewerConfig.registerConfig(emobj.ParticlesTiltPair,
                                   {ORDER: labels,
                                   VISIBLE: labels,
                                    MODE: MODE_MD,
                                    RENDER: renderLabels})

# SetOfClasses2D
labels = 'enabled id _size _representative._filename _filename'
RegistryViewerConfig.registerConfig(emobj.SetOfClasses2D,
                                   {ORDER: labels,
                                   VISIBLE: labels,
                                    RENDER: '_representative._filename',
                                    SORT_BY: '_size desc',
                                    LABELS: 'id _size'})



class DataViewer(pwviewer.Viewer):
    """ Wrapper to visualize different type of objects
    with the Xmipp program xmipp_showj
    """
    _name = "Xmipp metadata viewer"
    _environments = [pwviewer.DESKTOP_TKINTER, pwviewer.WEB_DJANGO]
    _targets = [
        emobj.Image,
        emobj.SetOfClasses2D,
        emobj.SetOfClasses3D,
        emobj.SetOfCoordinates,
        emobj.SetOfCTF,
        emobj.SetOfParticles,
        emobj.SetOfMovies,
        emobj.SetOfNormalModes,
        emobj.SetOfPrincipalComponents,
        emobj.SetOfPDBs,
        emobj.SetOfAtomStructs,
        emprot.ProtParticlePicking,
        emprot.ProtImportMovies,
        # TiltPairs related data
        emobj.CoordinatesTiltPair,
        emobj.MicrographsTiltPair,
        emobj.ParticlesTiltPair,
        emobj.SetOfImages,
    ]

    @classmethod
    def registerConfig(cls, type, config=None):
        """
        Allow registration of other objects
        :param type: Class to register --> SetOfXXX
        :param config: Optional, dictionary with the fields configuration, otherwise all fields will be shown.

        :return: Nothing

        """
        cls._targets.append(type)
        RegistryViewerConfig.registerConfig(type, config)

    def __init__(self, **kwargs):
        pwviewer.Viewer.__init__(self, **kwargs)
        self._views = []

    def _addObjView(self, obj, fn, viewParams={}):
        objView = ObjectView(self._project, obj.strId(), fn,
                             viewParams=viewParams)
        self._views.append(objView)
        return objView

    def _visualize(self, obj, **kwargs):
        cls = type(obj)

        if issubclass(cls, emobj.SetOfClasses2D):
            return [ClassesView(self._project, obj.strId(), obj.getFileName(), **kwargs)]


        # Try registry first
        config = RegistryViewerConfig.getConfig(cls)

        if config is not None:
            fn = obj.getFileName()
            objView = self._addObjView(obj, fn, config)
            if issubclass(cls, emobj.SetOfMovies):
                # For movies increase the JVM memory by 1 GB, just in case
                objView.setMemory(getJvmMaxMemory() + 1)
            return self._views

        if issubclass(cls, emobj.Volume):
            fn = emlib.image.ImageHandler.locationToXmipp(obj)

            if fn.endswith(':mrc'):
                # fn may came in the form of 000001@Runs/..../vol.mrc". We need to pass an actual file to Ccp4Header
                actualFile = fn.split("@")[-1]
                ccp4header = Ccp4Header(actualFile, readHeader=True)
                if ccp4header.getISPG() == 1:
                    fn = fn.replace(':mrc', '')
                else:
                    print('Developers warning: %s headers do not match the standard for volumes. '
                          'You can use fixVolume(pwem.convert.headers.py) '
                          'method to fix the headers' % fn)
            self._addObjView(obj, fn,
                             {RENDER: 'image',
                              SAMPLINGRATE: obj.getSamplingRate()})

        elif issubclass(cls, emobj.Image):
            fn = emlib.image.ImageHandler.locationToXmipp(obj)
            self._addObjView(obj, fn)

        # elif issubclass(cls, emobj.SetOfPDBs):
        #     fn = obj.getFileName()
        #     labels = 'id _filename '
        #     self._addObjView(obj, fn, {ORDER: labels,
        #                                VISIBLE: labels,
        #                                MODE: MODE_MD,
        #                                RENDER: "no"})

        # elif issubclass(cls, emobj.SetOfMovies):
        #     fn = obj.getFileName()
        #     # Enabled for the future has to be available
        #     labels = ('id _filename _samplingRate _acquisition._dosePerFrame '
        #               '_acquisition._doseInitial ')
        #     moviesView = self._addObjView(obj, fn, {ORDER: labels,
        #                                             VISIBLE: labels,
        #                                             MODE: MODE_MD,
        #                                             RENDER: "no"})
        #     # For movies increase the JVM memory by 1 GB, just in case
        #     moviesView.setMemory(getJvmMaxMemory() + 1)

        elif issubclass(cls, emobj.SetOfMicrographs):
            self._views.append(MicrographsView(self._project, obj, **kwargs))

        # elif issubclass(cls, emobj.MicrographsTiltPair):
        #     labels = 'id enabled _untilted._filename _tilted._filename'
        #     renderLabels = '_untilted._filename _tilted._filename'
        #     self._addObjView(obj, obj.getFileName(), {ORDER: labels,
        #                                               VISIBLE: labels,
        #                                               MODE: MODE_MD,
        #                                               RENDER: renderLabels})
        #
        # elif issubclass(cls, emobj.ParticlesTiltPair):
        #     labels = 'id enabled _untilted._filename _tilted._filename'
        #     renderLabels = '_untilted._filename _tilted._filename'
        #     self._addObjView(obj, obj.getFileName(), {ORDER: labels,
        #                                               VISIBLE: labels,
        #                                               RENDER: renderLabels,
        #                                               MODE: MODE_MD})

        elif issubclass(cls, emobj.SetOfCoordinates):
            # FIXME: Remove dependency on xmipp3 plugin to visualize coordinates
            xmipp3 = Domain.importFromPlugin('xmipp3',
                                             errorMsg="xmipp3 plugin is required "
                                                      "now to visualize coordinates.")
            micSet = obj.getMicrographs()  # accessing mics to provide metadata file
            if micSet is None:
                raise Exception('visualize: SetOfCoordinates has no micrographs set.')

            mdFn = getattr(micSet, '_xmippMd', None)
            if mdFn:
                fn = mdFn.get()
            else:  # happens if protocol is not an xmipp one
                fn = self._getTmpPath(micSet.getName() + '_micrographs.xmd')
                xmipp3.convert.writeSetOfMicrographs(micSet, fn)
            tmpDir = self._getTmpPath(obj.getName())
            pwutils.cleanPath(tmpDir)
            pwutils.makePath(tmpDir)
            # FIXME: (JMRT) We are always writing the SetOfCoordinates and removing
            # the tmpDir, we need to take into account if the user have pick
            # some particles in the tmpDir and have not save them, that now
            # will loose all picked particles.
            # A possible solution could be to alert that changes have not been
            # written during modification of tmpDir or create a new Xmipp picking
            # protocol to continue picking later without loosing the coordinates.
            xmipp3.convert.writeSetOfCoordinates(tmpDir, obj)
            self._views.append(CoordinatesObjectView(self._project, fn,
                                                     tmpDir, self.protocol,
                                                     inTmpFolder=True))

        # elif issubclass(cls, emobj.SetOfParticles):
        #     fn = obj.getFileName()
        #     labels = ('id enabled _index _filename _xmipp_zScore _xmipp_cumulativeSSNR '
        #               '_sampling _xmipp_scoreByVariance _xmipp_scoreEmptiness '
        #               '_ctfModel._defocusU _ctfModel._defocusV _ctfModel._defocusAngle '
        #               '_transform._matrix')
        #     self._addObjView(obj, fn, {ORDER: labels,
        #                                VISIBLE: labels,
        #                                SORT_BY: '_xmipp_zScore asc',
        #                                RENDER: '_filename'})

        # elif issubclass(cls, emobj.SetOfVolumes):
        #     fn = obj.getFileName()
        #     labels = 'id enabled comment _filename '
        #     self._addObjView(obj, fn, {MODE: MODE_MD,
        #                                ORDER: labels,
        #                                VISIBLE: labels,
        #                                RENDER: '_filename'})

        # elif issubclass(cls, emobj.SetOfClasses2D):
        #     self._views.append(ClassesView(self._project, obj.strId(),
        #                                    obj.getFileName(), **kwargs))
        #
        elif issubclass(cls, emobj.SetOfClasses3D):
            self._views.append(Classes3DView(self._project, obj.strId(),
                                             obj.getFileName()))

        elif issubclass(cls, emobj.SetOfImages):
            self._views.append(ObjectView(self._project, obj.strId(),
                                          obj.getFileName(), **kwargs))

        elif issubclass(cls, emobj.SetOfCTF):
            self._views.append(CtfView(self._project, obj))

        elif issubclass(cls, emobj.CoordinatesTiltPair):
            # FIXME: Remove dependency on xmipp3 plugin to visualize coordinates
            xmipp3 = Domain.importFromPlugin('xmipp3',
                                             errorMsg="xmipp3 plugin is required "
                                                      "now to visualize coordinates.")
            tmpDir = self._getTmpPath(obj.getName())
            pwutils.makePath(tmpDir)

            mdFn = os.path.join(tmpDir, 'input_micrographs.xmd')
            xmipp3.convert.writeSetOfMicrographsPairs(
                obj.getUntilted().getMicrographs(),
                obj.getTilted().getMicrographs(), mdFn)

            # TODO: Review this if ever a non Xmipp CoordinatesTiltPair is available
            xmipp3.convert.writeSetOfCoordinates(tmpDir, obj.getUntilted())
            xmipp3.convert.writeSetOfCoordinates(tmpDir, obj.getTilted())
            launchTiltPairPickerGUI(mdFn, tmpDir, self.protocol)

        elif issubclass(cls, emprot.ProtParticlePicking):
            if obj.getOutputsSize() >= 1:
                self._visualize(obj.getCoords())

        elif issubclass(cls, emprot.ProtImportMovies):
            movs = obj.outputMovies
            self._visualize(movs)
            gainFn = movs.getGain()
            if gainFn is not None:
                if os.path.exists(gainFn):
                    self._views.append(DataView(gainFn))

        elif issubclass(cls, emobj.SetOfNormalModes):
            self._views.append(DataView(obj.getFileName()))

        elif issubclass(cls, emobj.EMSet):
            self._views.append(DataView(obj.getFileName()))

        return self._views


class BasicMDViewer(DataViewer):
    """ Viewer to show sets in the metadataviewer to allow manual subsets when
    DataViewer shows it in a different way: e.g: Coordinates2D."""
    _name = "Basic MD viewer"
    _targets = [emobj.SetOfCoordinates, emobj.SetOfFSCs]

    def _visualize(self, obj, **kwargs):
        fn = obj.getFileName()
        self._addObjView(obj, fn, {MODE: MODE_MD})
        return self._views



# ***************************************************************************
# * Authors:    Roberto Marabini (roberto@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/
from tempfile import mkstemp

import numpy as np
import os

from pyworkflow.tests import BaseTest, setupTestProject
from pyworkflow.utils import runJob
from xmipp3 import Plugin
import pwem.protocols as emprot


def createFeatFile(fd):
    # create feat file
    f = os.fdopen(fd, "w")
    f.write("""# XMIPP_STAR_1 *
# Type of feature (sph, blo, gau, Cyl, dcy, cub, ell, con)(Required)
# The operation after adding the feature to the phantom (+/=) (Required)
# The feature density (Required)
# The feature center (Required)
# The vector for special parameters of each vector (Required)
# Sphere: [radius] Blob : [radius alpha m] Gaussian : [sigma]
# Cylinder : [xradius yradius height rot tilt psi]
# DCylinder : [radius height separation rot tilt psi]
# Cube : [xdim ydim zdim rot tilt psi]
# Ellipsoid : [xradius yradius zradius rot tilt psi]
# Cone : [radius height rot tilt psi]
data_block1
 _dimensions3D  '100 100 100'
 _phantomBGDensity  0.
 _scale  1.
data_block2
loop_
 _featureType
 _featureOperation
 _featureDensity
 _featureCenter
 _featureSpecificVector
""")
    theta_max = 8 * np.pi
    nPoints = 100
    thetaStep = theta_max/nPoints/1.2
    for theta in np.arange(-theta_max/2.4, theta_max/2.4, thetaStep):
        z = theta * 3.
        x = np.sin(theta) * theta_max*1.3
        y = np.cos(theta) * theta_max*1.3
        f.write("sph = 1 '%f %f %f' '5'\n" % (x, y, z))
    f.close()


def createParamFile(fd):
    f = os.fdopen(fd, "w")
    f.write("""# XMIPP_STAR_1 *
# Projection Parameters
#_angleFile
# X and Y projection dimensions [Xdim Ydim]
# Rotation range and number of samples [Start Finish NSamples]
# Rotation angle added noise  [noise (bias)]
# Type of range for Rotation (random_group or random or even)
# Tilt range and number of samples for Tilt
# Tilt angle added noise
# Type of range for Tilt
# Psi range and number of samples
# Psi added noise
# Type of range for Psi
# Noise applied to pixels [noise (bias)]
# Noise applied to particle center coordinates [noise (bias)]

data_block1
_dimensions2D   '100 100'
_projRotRange    '0 360 4'
_projRotNoise   '0.0 0.0'
_projRotRandomness   random
_projTiltRange    '0 180 4'
_projTiltNoise   '0.0 0.0'
_projTiltRandomness   random
_projPsiRange    '0 360 4'
_projPsiNoise   '0.0 0.0'
_projPsiRandomness   random
_noisePixelLevel   '0. 0.'
""")
    f.close()


def projectPhantom(featFileName, paramFileName, particlesFileName):
    args = "-i %s -o %s" % (featFileName, particlesFileName)
    args += " --params %s" % paramFileName
    runJob(None, "xmipp_phantom_project", args, env=Plugin.getEnviron())


class TestProtAssignAngles(BaseTest):
    @classmethod
    def setUpClass(cls):
        setupTestProject(cls)
        (fd, cls.featFileName) = mkstemp(suffix=".feat")
        createFeatFile(fd)
        (fd, cls.paramFileName) = mkstemp(suffix=".param")
        createParamFile(fd)
        (fd, cls.stackFileName) = mkstemp(suffix=".stk")
        cls.xmdFileName = cls.stackFileName.replace(".stk", ".xmd")
        projectPhantom(cls.featFileName, cls.paramFileName,
                       cls.stackFileName)
        (fd, cls.stackFileName2) = mkstemp(suffix=".stk")
        cls.xmdFileName2 = cls.stackFileName2.replace(".stk", ".xmd")
        projectPhantom(cls.featFileName, cls.paramFileName,
                       cls.stackFileName2)

    def test_00_basicAlignment(self):
        # import first set of particles
        prot1 = self.newProtocol(emprot.ProtImportParticles,
                                 importFrom=emprot.ProtImportParticles.IMPORT_FROM_XMIPP3,
                                 mdFile=self.xmdFileName,
                                 magnification=10000,
                                 samplingRate=1,
                                 haveDataBeenPhaseFlipped=False
                                 )
        prot1.setObjLabel('import Particles 1')
        self.launchProtocol(prot1)

        # import second set of particles
        prot2 = self.newProtocol(emprot.ProtImportParticles,
                                 importFrom=emprot.ProtImportParticles.IMPORT_FROM_XMIPP3,
                                 mdFile=self.xmdFileName2,
                                 magnification=10000,
                                 samplingRate=1,
                                 haveDataBeenPhaseFlipped=False
                                 )
        prot2.setObjLabel('import Particles 2')
        self.launchProtocol(prot2)

        # assign transformation from 1 set to second set
        prot3 = self.newProtocol(emprot.ProtAlignmentAssign,
                                 inputParticles=prot2.outputParticles,
                                 inputAlignment=prot1.outputParticles,
                                 )
        prot3.setObjLabel('assign angle 1 -> 2')
        self.launchProtocol(prot3)
        for part1, part2, part3 in emprot.izip(
                prot1.outputParticles,
                prot2.outputParticles,
                prot3.outputParticles):
            result = np.allclose(part1.getTransform().getMatrix(),
                                 part3.getTransform().getMatrix())
            self.assertTrue(result)
            result = np.allclose(part2.getTransform().getMatrix(),
                                 part3.getTransform().getMatrix())
            self.assertFalse(result)
        self.assertEqual(len(prot3.outputParticles),
                         len(prot1.outputParticles))



# **************************************************************************
# *
# * Authors:    Daniel Marchán (da.marchan@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import pyworkflow.protocol.params as params
from pwem.protocols import EMProtocol
from pyworkflow.object import Integer, Float, String




class ProtMathematicalOperator(EMProtocol):
    """
    Protocol to make mathematical operations on different inputs
    """
    PARAM1 = 'X1'
    PARAM2 = 'X2'
    _label = 'mathematical operator'

    def __init__(self, **args):
        EMProtocol.__init__(self, **args)

    def _defineParams(self, form):
        """attr2
        Defines the parameters the protocol form will show and its behaviour
        :param form:
        """
        form.addSection(label='Input')

        # Attribute X1
        form.addParam('input1', params.IntParam,
                      label='Input %s:' % self.PARAM1,
                      default=3,
                      allowsPointers=True,
                      help='Write the value you want to consider as %s or'
                             ' if it is a input pointer select from the list of available inputs.' % self.PARAM1)

        # Attribute X2
        form.addParam('input2', params.IntParam,
                      label='Input %s:' % self.PARAM2,
                      default=2,
                      allowsPointers=True,
                      help='Write the value you want to consider as %s or'
                             ' if it is a input pointer select from the list of available inputs.' % self.PARAM2)

        # Result
        form.addParam('expression', params.StringParam, label="Formula: ",
                      important=True,
                      default = "%s + (%s*3)" % (self.PARAM1, self.PARAM2),
                      help='Write the mathematical formula you want to calculate, e.g: (%s + %s) * %s * 0.9 \n'
                             'Please note that %s and %s would be replaced by the value you selected for each case.'
                             ' It is important to use upper cases when writing the expression.' % (self.PARAM1, self.PARAM2, self.PARAM1, self.PARAM1, self.PARAM2))

        form.addParam('typeResult', params.EnumParam, default=1,
                      choices=['Int', 'Float', 'String'],
                      label='Result type:',
                      help='Choose the variable type you want your result to be, e.g: Integer',
                      expertLevel=params.LEVEL_ADVANCED)

    def _insertAllSteps(self):
        self._insertFunctionStep(self.computeStep, self.getParam1(), self.getParam2() , prerequisites=[])

    def computeStep(self, attr1, attr2):
        """
        Applies the formula to each of the attributes or the numeric formula provided.
        Complex python code could be run separating lines with ;
        To use numpy you could do:
        import numpy;
        item._resolution.set(numpy.random.randint(10))
        """
        str1 = str(attr1)
        str2 = str(attr2)
        strInit = self.expression.get()
        strInit = strInit.replace(self.PARAM1, str1)
        self.formula = strInit.replace(self.PARAM2, str2)
        self.info("FORMULA after replacement: %s" % self.formula)
        result = eval(self.formula)
        self.info("RESULT: %s" % result )
        self.createResultOutput(result)

        self.info('To calculate: %s' % self.formula)
        self.info('Result = %s' % result)

    def createResultOutput(self, result):
        """ The output can be an Integer, Float or String. Other protocols can use it in those
            Params if it has set allowsPointer=True
        """

        tpR = self.typeResult.get()
        if tpR == 0:
            resultNum = Integer(result)
        elif tpR == 1:
            resultNum = Float(result)
        elif tpR == 2:
            resultNum = String(result)

        self._defineOutputs(result=resultNum)

    def formulaNeedsParam(self, param):
        """ return True if the formula uses the param passed"""

        return param in self.expression.get()

    def formulaNeedsInput(self):
        """ Returns true if there is need for any input"""
        return self.formulaNeedsParam(self.PARAM1) or self.formulaNeedsParam(self.PARAM2)

    def _summary(self):
        summary = []

        summary.append('To calculate: %s' % self.expression.get())
        if hasattr(self, "formula"):
            summary.append('To calculate: %s' % self.formula)
        if hasattr(self, "result"):
            summary.append('Result = %s' % self.result)

        return summary

    def _validate(self):
        errors = []
        if self.formulaNeedsParam(self.PARAM1):
            if self.getParam1() is None:
                errors.append('Input %s should be provided.' % self.PARAM1)

        if self.formulaNeedsParam(self.PARAM2):
            if self.getParam2() is None:
                errors.append('Input %s should be provided.' % self.PARAM2)

        return errors


    def getParam2(self):

        return self.input2.get()

    def getParam1(self):

        return self.input1.get()


# -*- coding: utf-8 -*-
# **************************************************************************
# *
# * Authors:     Marta Martinez (mmmtnez@cnb.csic.es)
# *              Roberto Marabini (roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
from urllib.error import URLError, HTTPError
from urllib.request import urlopen
from os.path import exists


# sequence related stuff


from Bio.Seq import Seq
from Bio import Entrez, SeqIO
import sys
from Bio.SeqRecord import SeqRecord
from Bio.Align.Applications import ClustalOmegaCommandline, MuscleCommandline
from Bio import pairwise2
from pyworkflow.utils import getExt
from pwem.objects.data import Alphabet


class SequenceHandler:
    def __init__(self, sequence=None,
                 iUPACAlphabet=Alphabet.DUMMY_ALPHABET, doClean=True):
        iUPACAlphabet = Alphabet.DUMMY_ALPHABET if iUPACAlphabet not in Alphabet.alphabets else iUPACAlphabet

        if iUPACAlphabet == Alphabet.DUMMY_ALPHABET:
            self.isAminoacid = None
        else:      
            self.isAminoacid =  (iUPACAlphabet <   Alphabet.AMBIGOUS_DNA_ALPHABET)

        self.alphabet = iUPACAlphabet ##  indexToAlphabet(isAminoacid, iUPACAlphabet)

        if sequence is not None:
            if doClean:
                self._sequence = cleanSequence(self.alphabet, sequence)
            else:
                self._sequence = sequence
        else:
            self._sequence = None

    def getSequence(self):
        """Returns the sequence as Bio.Seq.Seq object"""
        return Seq(self._sequence)

    def getTypeFromFile(self, fileName):
        '''Returns the expected BioPython file type according to the filename'''
        ext = getExt(fileName)
        if ext == '.fasta':
            type = 'fasta'
        elif ext == '.genbank' or ext == '.gb':
            type = 'genbank'
        else:
            type = 'fasta'
        return type

    def appendFile(self, fileName, seqID, sequence=None, name=None,
                 seqDescription=None, type="fasta"):
        '''Appends a sequence to the sequences in a file'''
        if sequence is not None:
            self._sequence = sequence

        if type is None:
            type = self.getTypeFromFile(fileName)

        if exists(fileName):
            records = list(SeqIO.parse(fileName, type))
        else:
            records = []
        records.append(SeqRecord(Seq(self._sequence), id=seqID, name=name,
                                 description=seqDescription))
        with open(fileName, "w") as output_handle:
            SeqIO.write(records, output_handle, type)

    def saveFile(self, fileName, seqID, sequence=None, name=None,
                 seqDescription=None, type="fasta"):
        '''Saves a single sequence into a specified file'''
        if sequence is not None:
            self._sequence = sequence

        self.saveFiles(fileName, [seqID], [self._sequence], [name], [seqDescription], type)

    def saveFiles(self, fileName, seqIDs, sequences, names,
                  seqDescriptions, type="fasta"):
        '''Saves multiple sequences into a specified file'''
        if type is None:
            type = self.getTypeFromFile(fileName)

        records = []
        for seq, seqID, name, seqDescription in zip(sequences, seqIDs, names, seqDescriptions):
            records.append(SeqRecord(Seq(seq), id=seqID, name=name,
                               description=seqDescription))
        # type(record): < class 'Bio.SeqRecord.SeqRecord'>
        with open(fileName, "w") as output_handle:
            SeqIO.write(records, output_handle, type)

    def readSequenceFromFile(self, fileName, type="fasta", isAmino=True):
        '''From a sequences file, returns a dictionary with ther FIRST sequence info.
        Dictionary: {'seqID': seqID1, 'sequence': sequence1, 'description': description1, 'alphabet': alphabet1}'''

        if type is None:
            type = self.getTypeFromFile(fileName)
        return self.readSequencesFromFile(fileName, isAmino=isAmino)[0]

    def readSequencesFromFile(self, fileName, type='fasta', isAmino=True):
        '''From a sequences file, returns a list of dictionaries with each sequence info.
        Dictionary: [{'seqID': seqID1, 'sequence': sequence1, 'description': description1, 'alphabet': alphabet1},
                     ...]'''

        if type is None:
            type = self.getTypeFromFile(fileName)

        sequences = []

        records = SeqIO.parse(fileName, type)

        for rec in records:
            sequence = str(rec.seq)

            if isAmino:
                self.alphabet = Alphabet.EXTENDED_PROTEIN_ALPHABET
            elif 'U' in sequence:
                self.alphabet = Alphabet.UNAMBIGOUS_RNA_ALPHABET
            elif 'T' in sequence:
                self.alphabet = Alphabet.UNAMBIGOUS_DNA_ALPHABET
            else:
                self.alphabet = Alphabet.NUCLEOTIDES_ALPHABET

            sequences.append({'seqID': rec.id, 'sequence': sequence,
                              'name': rec.name, 'description': rec.description,
                              'isAminoacids': isAmino, 'alphabet': self.alphabet})
        return sequences

    def downloadSeqFromDatabase(self, seqID, dataBase=None):
        # see http://biopython.org/DIST/docs/api/Bio.SeqIO-module.html
        # for format/databases

        seqID = str(seqID)
        sys.stdout.flush()
        counter = 1
        retries = 5
        record = None
        error = ""
        if dataBase is None:
            if self.isAminoacid:
                dataBase = 'UniProt'
            else:
                dataBase = 'GeneBank'

        while counter <= retries:  # retry up to 5 times if server busy
            try:
                if dataBase == 'UniProt':
                    url = "http://www.uniprot.org/uniprot/%s.xml"
                    format = "uniprot-xml"

                    handle = urlopen(url % seqID)
                    alphabet = Alphabet.EXTENDED_PROTEIN_ALPHABET
                else:
                    if self.isAminoacid:
                        db = "protein"
                        alphabet = Alphabet.EXTENDED_PROTEIN_ALPHABET
                    else:
                        db = "nucleotide"
                        alphabet = Alphabet.NUCLEOTIDES_ALPHABET
                    Entrez.email = "scipion@cnb.csic.es"
                    format = "fasta"
                    handle = Entrez.efetch(db=db, id=seqID,
                                           rettype=format, retmode="text")

                record = SeqIO.read(handle, format)
                break
            except HTTPError as e:
                error = "%s is a wrong sequence ID" % seqID
                print(e.code)
            except URLError as e:
                error = "Cannot connect to %s" % dataBase
                print(e.args)
            except Exception as ex:
                template = "An exception of type {0} occurred. Arguments:\n{1!r}"
                message = template.format(type(ex).__name__, ex.args)
                error = message
            if counter == retries:
                break
            counter += 1

        seqDic = None
        if record is not None:
            seqDic = {'seqID': record.id, 'sequence': str(record.seq), 'description': record.description,
                      'alphabet': alphabet}
        return seqDic, error

    def alignSeq(self, referenceSeq):
        if self._sequence is not None:
            alignments = pairwise2.align.globalds(self._sequence.seq,
                                                  referenceSeq.seq)
            return alignments
        else:
            print("read the sequence first")
            exit(0)


def sequenceLength(filename, format='fasta'):
    handler = SequenceHandler()
    return len(handler.readSequenceFromFile(filename, format)['sequence'])


def cleanSequenceScipion(isAminoacid, iUPACAlphabet, sequence):
    return cleanSequence(iUPACAlphabet, sequence)


def cleanSequence(alphabet, sequence):
    """Remove all characters that are not in the alphabet
       :param alphabet: the alphabet to use as integer
       :param sequence: the sequence to clean
       """
    str_list = []
    for item in sequence.upper():
        if item in Alphabet.alphabets[alphabet]:
            str_list.append(item)
    value = ''.join(str_list)
    return ''.join(str_list)




def saveFileSequencesToAlign(SeqDic, inFile, type="fasta"):
    # Write my sequences to a fasta file
    with open(inFile, "w") as output_handle:
        for index, seq in SeqDic.items():
            record = SeqRecord(seq, id=str(index),
                               name="", description="")
            SeqIO.write(record, output_handle, type)


def alignClustalSequences(inFile, outFile):
    # Alignment of sequences with Clustal Omega program
    clustalomega_cline = ClustalOmegaCommandline(
            infile=inFile,
            outfile=outFile,
            verbose=True, auto=True)
    return clustalomega_cline


def alignMuscleSequences(inFile, outFile):
    # Alignment of sequences with Muscle program
    muscle_cline = MuscleCommandline(input=inFile, out=outFile)
    return muscle_cline


def alignBioPairwise2Sequences(structureSequenceId, structureSequence,
                               userSequenceId, userSequence,
                               outFileName):
    """aligns two sequences and saves them to disk using fasta format"""
    # see alignment_function for globalms parameters
    alignments = pairwise2.align.globalms(structureSequence,
                                          userSequence,  3, -1, -3, -2)
    align1, align2, score, begin, end = alignments[0]
    with open(outFileName, "w") as handle:
        handle.write(">%s\n%s\n>%s\n%s\n" % (structureSequenceId,
                                             align1,
                                             userSequenceId,
                                             align2))



# **************************************************************************
# *
# * Authors:     Marta Martinez (mmmtnez@cnb.csic.es)
# *              Roberto Marabini (roberto@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************


"""
This module implements visualization program
for input volumes.
"""

import os
from shutil import which
from tkinter.messagebox import showerror

import pyworkflow.protocol.params as params
import pyworkflow.viewer as pwviewer

import pwem.objects as emobj
import pwem.protocols as emprot
from pwem.viewers import Chimera, ChimeraView, EmProtocolViewer
from pwem.emlib import image

VOLUME_SLICES = 1
VOLUME_CHIMERA = 0


class viewerProtImportVolumes(EmProtocolViewer):
    """ Wrapper to visualize different type of objects
    with the Xmipp program xmipp_showj. """

    _label = 'viewer input volume'
    _targets = [emprot.ProtImportVolumes, emprot.ProtOrigSampling]
    _environments = [pwviewer.DESKTOP_TKINTER, pwviewer.WEB_DJANGO]

    def _defineParams(self, form):
        form.addSection(label='Visualization of input volumes')
        form.addParam('displayVol', params.EnumParam,
                      choices=['chimerax', 'slices'],
                      default=VOLUME_CHIMERA,
                      display=params.EnumParam.DISPLAY_HLIST,
                      label='Display volume with',
                      help='*chimerax*: display volumes as surfaces with '
                           'ChimeraX.\n *slices*: display volumes as 2D slices '
                           'along z axis.\n If the number of volumes is equal to 1, '
                           'a system of coordinates is shown'
                      )

    def _getVisualizeDict(self):
        return {
            'displayVol': self._showVolumes,
        }

    def _validate(self):
        if (self.displayVol == VOLUME_CHIMERA
                and which(Chimera.getProgram()) is None):
            return ["chimera is not available. "
                    "Either install it or choose option 'slices'. "]
        return []

    # =========================================================================
    # ShowVolumes
    # =========================================================================

    def _showVolumes(self, paramName=None):
        if self.displayVol == VOLUME_CHIMERA:
            return self._showVolumesChimera()

        elif self.displayVol == VOLUME_SLICES:
            return self._showVolumesSlices()

    def _createSetOfVolumes(self):
        try:
            setOfVolumes = self.protocol.outputVolumes
            sampling = self.protocol.outputVolumes.getSamplingRate()
        except Exception as ex:
            setOfVolumes = self.protocol._createSetOfVolumes()
            setOfVolumes.append(self.protocol.outputVolume)
            sampling = self.protocol.outputVolume.getSamplingRate()

        return sampling, setOfVolumes

    def _showVolumesChimera(self):
        """ Create a chimera script to visualize selected volumes. """
        tmpFileNameCMD = self.protocol._getExtraPath("chimera.cxc")
        f = open(tmpFileNameCMD, "w")
        sampling, _setOfVolumes = self._createSetOfVolumes()
        count = 1  # chimeraX stars counting in 1 (chimera in 0)

        if len(_setOfVolumes) == 1:
            # if we have a single volume then create axis
            # as bild file. Chimera must read the bild file first
            # otherwise system of coordinates will not
            # be in the center of the window

            dim = self.protocol.outputVolume.getDim()[0]
            tmpFileNameBILD = os.path.abspath(self.protocol._getExtraPath(
                "axis.bild"))
            Chimera.createCoordinateAxisFile(dim,
                                             bildFileName=tmpFileNameBILD,
                                             sampling=sampling)
            f.write("open %s\n" % tmpFileNameBILD)
            f.write("cofr 0,0,0\n")  # set center of coordinates
            count = 2  # skip first model because is not a 3D map

        for vol in _setOfVolumes:
            localVol = os.path.abspath(image.ImageHandler.removeFileType(
                vol.getFileName()))
            if localVol.endswith("stk"):
                errorWindow(None, "Extension .stk is not supported")
            f.write("open %s\n" % localVol)
            f.write("volume #%d style surface level 0.001 voxelSize %f\n" %
                    (count, sampling))
            count += 1

        if len(_setOfVolumes) > 1:
            f.write('tile\n')
        else:
            x, y, z = vol.getShiftsFromOrigin()
            f.write("volume #2 origin %0.2f,%0.2f,%0.2f\n" % (x, y, z))
            if vol.getHalfMaps():
                for halfMap in vol.getHalfMaps().split(','):
                    if not os.path.abspath(halfMap).endswith(".mrc"):
                        f.write("open %s\n" % (os.path.abspath(halfMap).split(".")[0] + ".mrc"))
                    else:
                        f.write("open %s\n" % os.path.abspath(halfMap))
                    f.write("volume #%d style surface level 0.001 voxelSize %f\n" %
                            (count, sampling))
                    f.write("volume #%d origin %0.2f,%0.2f,%0.2f\n" %
                            (count, x, y, z))
                    f.write("tile\n")
                    count += 1
        f.write("view\n")
        f.close()
        return [ChimeraView(tmpFileNameCMD)]

    def _showVolumesSlices(self):
        # Write an sqlite with all volumes selected for visualization.
        sampling, setOfVolumes = self._createSetOfVolumes()
        # setOfVolumes.setSamplingRate(sampling)
        if len(setOfVolumes) == 1:
            if self.protocol.outputVolume.getHalfMaps():
                setOfVolumes = self.protocol._createSetOfVolumes(suffix='tmp')
                vol = emobj.Volume()
                vol.setFileName(self.protocol.outputVolume.getFileName())
                vol.setSamplingRate(sampling)
                setOfVolumes.append(vol)
                for halfMap in self.protocol.outputVolume.getHalfMaps().split(','):
                    volHalf = emobj.Volume()
                    volHalf.setSamplingRate(sampling)
                    if not halfMap.endswith(".mrc"):
                        volHalf.setFileName(halfMap.split(".")[0] + ".mrc")
                    else:
                        volHalf.setFileName(halfMap)
                    setOfVolumes.append(volHalf)
                setOfVolumes.write()
                return [self.objectView(setOfVolumes)]
            else:
                return [self.objectView(self.protocol.outputVolume)]
        return [self.objectView(setOfVolumes)]


def errorWindow(tkParent, msg):
    try:
        showerror("Error",  # bar title
                  msg,  # message
                  parent=tkParent)
    except Exception as ex:
        print("Error:", msg)




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import pyworkflow.protocol.params as params
from pyworkflow.protocol import Protocol


class ProtTestParallel(Protocol):
    """ A parallel test protocol.
    """    
    _label = "parallel test"
    
    def __init__(self, **args):
        Protocol.__init__(self, **args)        
        self.stepsExecutionMode = params.STEPS_PARALLEL
        
    # --------------------------- DEFINE param functions ----------------------
    def _defineParams(self, form):
        form.addSection(label='Input')
        form.addParam('numberOfIterations', params.IntParam, default=2,
                      label="Number of iterations", 
                      help='Repeat the insertion of steps N times.')
        form.addParam('numberOfParallelSleeps', params.IntParam, default=2,
                      label="Number of parallel sleeps", 
                      help='How many sleep steps can be done at the same time.')
        form.addParam('failAfter', params.IntParam, default=0,
                      label="Fail after", 
                      help='If you set an id, the next step should fail')
        form.addParam('sleepSecs', params.IntParam, default=2,
                      label='Seconds to sleep')
        
        form.addParallelSection(threads=4, mpi=1)
            
    # --------------------------- INSERT steps functions ----------------------
    def _insertAllSteps(self):
        n = self.numberOfIterations.get()
        m = self.numberOfParallelSleeps.get()
        secs = self.sleepSecs.get()
        failAfter = self.failAfter.get()
        
        deps = []
        
        for i in range(n):
            initId = self._insertFunctionStep('initStep', i, prerequisites=deps)
            deps = []
            for j in range(m):
                fail = deps[-1] == failAfter if deps else False
                tag = 'iter %d, n: %d' % (i, j)
                sleepId = self._insertFunctionStep('sleepStep', secs, fail, tag,
                                                   prerequisites=[initId])
                deps.append(sleepId)
            endId = self._insertFunctionStep('endStep', i, prerequisites=deps)
            deps = [endId]
            
    # --------------------------- STEPS functions -----------------------------
    def initStep(self, iterN):
        """ All subsequent sleep steps should depend on this. """
        self._log.info("Starting iteration: %d" % iterN)
    
    def sleepStep(self, secs=5, forceFail=False, tag=''):
        if forceFail:
            self.runJob('echo', " 'Failing for testing purposes...'; exit 1")
        else:
            from pwem import Domain
            getEnviron = Domain.importFromPlugin('xmipp3', 'getEnviron')
            self.runJob('xmipp_work_test',
                        "--time %d --tag '%s'" % (secs, tag), env=getEnviron())
            
    def awakeStep(self):
        print("Awakened after a sleep step")
        
    def endStep(self, iterN):
        self._log.info("Ending iteration: %d" % iterN)



# **************************************************************************
# *
# * Authors:     Airen Zaldivar Peraza (azaldivar@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
"""
In this module are protocol base classes related to 2D processing

"""
import pyworkflow.protocol.params as params
import pyworkflow.utils as pwutils

from pwem.protocols import EMProtocol


class Prot3D(EMProtocol):
    pass


class ProtPreprocessVolumes(Prot3D):
    """ This class will serve as a base for all protocol
    that performs some operation on Volumes (i.e. filters, mask, resize, etc)
    It is mainly defined by an inputVolumes and outputVolumes.
    """
    def _defineParams(self, form):
        form.addSection(label=pwutils.Message.LABEL_INPUT)
        
        form.addParam('inputVolumes', params.PointerParam, important=True,
                      label=pwutils.Message.LABEL_INPUT_VOLS, pointerClass='Volume, SetOfVolumes',
                      help='Can be a density volume or a SetOfVolumes.')
        # Hook that should be implemented in subclasses
        self._defineProcessParams(form)
        form.addParallelSection(threads=2, mpi=1)
        
    def _defineProcessParams(self, form):
        """ This method should be implemented by subclasses
        to add other parameter relatives to the specific operation."""
        pass


class ProtFilterVolumes(ProtPreprocessVolumes):
    """ This is the base for the branch of filters, 
    between the ProtPreprocessVolumes """
    pass


class ProtOperateVolumes(ProtPreprocessVolumes):
    """ This is the base for the branch of filters,
    between the ProtPreprocessParticles """
    def __init__(self, **args):
        ProtPreprocessVolumes.__init__(self, **args)


class ProtMaskVolumes(ProtPreprocessVolumes):
    """ This is the base for the branch of mask, 
    between the ProtPreprocessVolumes """
    pass


class ProtCreateMask3D(ProtPreprocessVolumes):
    pass


class ProtAlignVolume(ProtPreprocessVolumes):
    """Protocol base for Align volumes protocols"""
    pass


class ProtReconstruct3D(Prot3D):
    pass


class ProtRefine3D(Prot3D):
    pass


class ProtClassify3D(Prot3D):
    pass


class ProtInitialVolume(Prot3D):
    """Protocol base for Initial volumes protocols"""
    pass


class ProtAnalysis3D(Prot3D):
    pass


class ProtFitting3D(Prot3D):
    pass



# ***************************************************************************
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/


import pyworkflow.tests as pwtests

import pwem.protocols as emprot


class TestImportBase(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dsXmipp = pwtests.DataSet.getDataSet('xmipp_tutorial')

    
class TestImportMasks(TestImportBase):
    
    def test_import_mask2d(self):
        """ Import a mask 2d.
        """
        args = {'maskPath': self.dsXmipp.getFile('mask2d'),
                'samplingRate': 2.1
                }

        prot = self.newProtocol(emprot.ProtImportMask, **args)
        prot.setObjLabel('import mask 2d')
        self.launchProtocol(prot)
        self.assertIsNotNone(prot.outputMask,
                             "There was a problem when importing a 2d mask.")

    def test_import_mask3d(self):
        """ Import a mask 3d.
        """
        args = {'maskPath': self.dsXmipp.getFile('mask3d'),
                'samplingRate': 2.1
                }

        prot = self.newProtocol(emprot.ProtImportMask, **args)
        prot.setObjLabel('import mask 3d')
        self.launchProtocol(prot)
        self.assertIsNotNone(prot.outputMask,
                             "There was a problem when importing a 3d mask.")




# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (delarosatrevin@scilifelab.se) [1]
# *
# * [1] SciLifeLab, Stockholm University
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os

import pyworkflow.utils as pwutils
import pyworkflow.viewer as pwviewer

import pwem.objects as emobj


class Vmd:
    """ Help class to run VMD and manage its environment. """

    @classmethod
    def getEnviron(cls):
        """ Return the proper environ to launch VMD.
        VMD_HOME variable is read from the ~/.config/scipion.conf file.
        """
        environ = pwutils.Environ(os.environ)
        environ.set('PATH', os.path.join(os.environ['VMD_HOME'], 'bin'),
                    position=pwutils.Environ.BEGIN)
        return environ


class VmdView(pwviewer.CommandView):
    """ View for calling an external command. """
    def __init__(self, vmdCommand, **kwargs):
        pwviewer.CommandView.__init__(self, 'vmd %s' % vmdCommand,
                                      env=Vmd.getEnviron(), **kwargs)

    def show(self):
        pwutils.runJob(None, '', self._cmd, env=Vmd.getEnviron())


class VmdViewer(pwviewer.Viewer):
    """ Wrapper to visualize PDB objects with VMD viewer. """
    _environments = [pwviewer.DESKTOP_TKINTER]
    # _targets = [AtomStruct]

    def __init__(self, **args):
        pwviewer.Viewer.__init__(self, **args)

    def visualize(self, obj, **args):
        cls = type(obj)

        if issubclass(cls, emobj.AtomStruct):
            VmdView(obj.getFileName()).show()
            # FIXME: there is an asymmetry between ProtocolViewer and Viewer.
            # For the first, the visualize method return a list of View's,
            # while for the second, the visualize method directly shows
            # the objects. (the first approach is preferable)
        else:
            raise Exception('VmdViewer.visualize: can not visualize class: %s'
                            % obj.getClassName())



# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os
try:
    from itertools import izip
except ImportError:
    izip = zip

from json import dumps, loads

from pyworkflow.protocol.params import (PointerParam, FileParam, StringParam)
from pyworkflow.utils import moveFile

from pwem.protocols import EMProtocol
import pwem.objects as emobj
from pwem.constants import ALIGN_NONE
from pwem.objects import (MicrographsTiltPair, ParticlesTiltPair)


class BatchProtocol(EMProtocol):
    """ Base class to all protocols that are launched
    through other GUIs (such as showj) and that
    are called "batch" protocols. They should not be
    executed from normal "form" of other protocols.
    """
    pass


class ProtUserSubSet(BatchProtocol):
    """ Create subsets from the GUI.
    This protocol will be executed mainly from the ShowJ gui or the new metadata viewer.
    For ShowJ: The enabled/disabled changes will be stored in a temporary sqlite
    For the Metadata viewer: the original sqlite is passed and a path of a txt file with the ids selected.

    Example of a ShowJ socket message:

     run protocol ProtUserSubSet inputObject=380 sqliteFile='Runs/000335_ProtImodTSNormalization/tiltseries_state.sqlite','' outputClassName=SetOfTiltSeries other='' label='create subset'

     run protocol is received by ProjectTCPRequestHandler.handle()

    """
    def __init__(self, **args):
        BatchProtocol.__init__(self, **args)
        self._selectedIds = None
        self._dbName = None # To keep the path to the database to read
        self._selectionTxt = None # To keep the file (txt) that has the selection. Extracted from sqliteFile tuple.

    def _defineParams(self, form):
        form.addSection("Input")
        form.addParam('inputObject', PointerParam, label="Input set", pointerClass='EMSet')
        form.addParam('other', StringParam, allowsNull=True, label="Other")
        form.addParam('sqliteFile', FileParam, label="Selection file")
        form.addParam('outputClassName', StringParam, label="Output type")

    def _insertAllSteps(self):
        self._insertFunctionStep('createSetStep')

    def createSetStep(self):

        sourceSet = self.inputObject.get()
        markedSet = self.createSetObject()  # Set equal to sourceSet but marked with disabled
        other = self.other.get()

        self.info("Source: %s" % sourceSet) # Scipion object input for this protocol
        self.info("Output type: %s" % self.outputClassName)
        self.info("Subset file: %s" % self.sqliteFile) # Sqlite with the selection (ShowJ case) or
        if other:
            self.info("Other: %s" % other)


        # New recommended way to create subsets: making the set responsible for his own subset process
        # Once all Sets implement appendFromSet and the if below is gone we can remove this "if"
        if getattr(markedSet, "USE_CREATE_COPY_FOR_SUBSET", False):
            markedSet.loadAllProperties()
            newSet = markedSet.createCopy(self._getPath(),
                                          copyInfo=True, copyItems=True)

            # Define outputs, may be use something more specific than "subset"
            self._defineOutputs(subset=newSet)
            self._defineSourceRelation(sourceSet, newSet)
            return

        if other and ',Volume' in other:
            volId = int(other.split(',')[0])

            if isinstance(markedSet, emobj.SetOfVolumes):
                volSet = emobj.SetOfVolumes(filename=self._dbName)
                output = volSet[volId]
            else:
                classSet = emobj.SetOfClasses3D(filename=self._dbName)
                output = classSet[volId].getRepresentative()
            self._defineOutputs(outputVolume=output)

        elif isinstance(sourceSet, emobj.SetOfImages):
            self._createSubSetFromImages(sourceSet)

        elif isinstance(sourceSet, emobj.SetOfClasses):
            self._createSubSetFromClasses(sourceSet)

        elif isinstance(sourceSet, emobj.SetOfCTF):
            outputClassName = self.outputClassName.get()
            if outputClassName.startswith('SetOfMicrographs'):
                self._createMicsSubSetFromCTF(sourceSet)

        elif isinstance(sourceSet, emobj.SetOfAtomStructs):
            self._createSubSetFromAtomStructs(sourceSet)

        elif isinstance(sourceSet, MicrographsTiltPair):
            self._createSubSetFromMicrographsTiltPair(sourceSet)

        elif isinstance(sourceSet, ParticlesTiltPair):
            self._createSubSetFromParticlesTiltPair(sourceSet)

        elif isinstance(sourceSet, EMProtocol):
            if self.other.hasValue():
                otherid = self.other.get()
                otherObj = self.getProject().mapper.selectById(int(otherid))

                if isinstance(markedSet, emobj.SetOfClasses):
                    markedSet.setImages(otherObj)
                    self._createSubSetFromClasses(markedSet)

                elif isinstance(markedSet, emobj.SetOfImages):
                    markedSet.copyInfo(otherObj)  # copy info from original images
                    self._createSubSetFromImages(markedSet)

                elif isinstance(markedSet, emobj.SetOfNormalModes):
                    self._createSimpleSubset(otherObj)
            else:
                if isinstance(markedSet, emobj.SetOfVolumes):
                    volSet = emobj.SetOfVolumes(filename=self._dbName)
                    volSet.loadAllProperties()
                    self._createSimpleSubset(volSet)

                # Go for a generic way of creating the set the the
                # input set is not registered (typically from viewers)
                else:
                    # We might want to do this before, inside the createSetObject
                    markedSet.loadAllProperties()
                    self._createSimpleSubset(markedSet)

        else:
            self._createSimpleSubset(sourceSet)

    def _createSimpleSubset(self, inputObj):
        className = inputObj.getClassName()
        modifiedSet = inputObj.getClass()(filename=self._dbName,
                                          prefix=self._dbPrefix)
        try:
            createFunc = getattr(self, '_create' + className)
            output = createFunc()
        except Exception:
            output = inputObj.createCopy(self._getPath())

        for item in modifiedSet:
            if self._itemSelected(item):
                output.append(item)

        if hasattr(modifiedSet, 'copyInfo'):
            output.copyInfo(inputObj)

        # Register outputs
        self._defineOutput(className, output)

        if inputObj.hasObjId():
            self._defineTransformRelation(inputObj, output)

        return output

    def usingShowJ(self):
        return ".sqlite" in self.sqliteFile.get()

    def _itemSelected(self, item):
        """ Returns true if the element is selected. We will have to modes: ShowJ or Metadata viewer
        ShowJ marks the items as enable=False
        Metadata viewer, passes a file with the id selected

        """
        if self.usingShowJ():
            return item.isEnabled()
        else:
            return self._itemInSelectionTxt(item)

    def _itemInSelectionTxt(self, item):

        id = item.getObjId()
        return id in self._getSelectionTxtDict()

    def _createSubSetFromImages(self, inputImages,
                                copyInfoCallback=None):
        className = inputImages.getClassName()
        setClass = inputImages.getClass()

        modifiedSet = setClass(filename=self._dbName, prefix=self._dbPrefix)
        try:
            createFunc = getattr(self, '_create' + className)
            output = createFunc()
        except Exception:
            output = inputImages.createCopy(self._getPath())

        if copyInfoCallback is None:
            # modifiedSet.loadAllProperties()
            output.copyInfo(inputImages)
        else:
            copyInfoCallback(output)

        output.appendFromImages(modifiedSet, self._itemSelected)
        # Register outputs
        self._defineOutput(className, output)

        if inputImages.hasObjId():
            self._defineTransformRelation(inputImages, output)

        # Define an informative summary of the subset operation
        sizeIn = inputImages.getSize()
        sizeOut = output.getSize()
        sizeDiff = sizeIn - sizeOut
        msg = 'A subset of _%s_ was created, ' % output.getClassName()
        msg += ('discarding *%d* items (%0.1f %%) from the input set.' %
                (sizeDiff, sizeDiff*100./sizeIn))
        self.summaryVar.set(msg)

        return output

    def _getSelectionTxtDict(self):

        if self._selectedIds is None:
            self._selectedIds = dict()
            ids = None
            self.info("Reading selection from %s" % self._dbName)
            # This file has a single line with id separated by spaces
            with open(self._selectionTxt, "r") as fh:
                line = fh.readline().strip()
                ids = line.split(" ")

            for id in ids:
                self._selectedIds[int(id)] = id

        return self._selectedIds

    def _createSubSetFromClasses(self, inputClasses):
        outputClassName = self.outputClassName.get()

        if (outputClassName.startswith('SetOfAverages') or
            outputClassName.startswith('SetOfVolumes') or
                outputClassName.startswith('SetOfParticles')):
            # We need to distinguish two cases:
            # a) when we want to create images by grouping class images
            # b) create a subset from a particular class images
            from pyworkflow.mapper.sqlite import SqliteFlatDb
            db = SqliteFlatDb(dbName=self._dbName, tablePrefix=self._dbPrefix)
            itemClassName = db.getSelfClassName()

            if itemClassName.startswith('Class'):

                if outputClassName.startswith('SetOfParticles'):
                    # Just to be sure, we check first if we want to create a subset of the items of the class
                    # If this check is not done in first place, it might be possible to enter through the second
                    # check (if REP_SET_TYPE is set), generating a SetOfVolumes like instead of a SetOfParticles like
                    # object
                    return self._createImagesFromClasses(inputClasses)

                elif hasattr(inputClasses, "REP_SET_TYPE"):
                    # Second check determines if we want to create a subset of representatives. In addition, we consider
                    # that the SetOfClasses stores the information about the type of set to be created
                    return self._createRepresentativesFromClasses(inputClasses,
                                                                  getattr(inputClasses, "REP_SET_TYPE"))
                else:
                    # Third check determines if we want to create a subset of representatives. By default,
                    # a SetOfVolumes is generated
                    return self._createRepresentativesFromClasses(inputClasses,
                                                                  outputClassName.split(',')[0])
            else:
                def callback(output):
                    self._copyInfoAndSetAlignment(inputClasses, output)

                return self._createSubSetFromImages(inputClasses.getImages(),
                                                    copyInfoCallback=callback)

        elif outputClassName.startswith('SetOfClasses'):
            return self._createClassesFromClasses(inputClasses)
        else:
            raise Exception("Unrecognized output type: '%s'" % outputClassName)

    def _createMicsSubSetFromCTF(self, inputCTFs):
        """ Create a subset of Micrographs and CTFs when analyzing the CTFs. """
        outputMics = self._createSetOfMicrographs()
        outputCtfs = self._createSetOfCTF()
        setOfMics = inputCTFs.getMicrographs()
        if setOfMics is None:
            raise Exception('Could not create SetOfMicrographs subset from '
                            'this SetOfCTF, the micrographs were not set.')
        outputMics.copyInfo(setOfMics)

        modifiedSet = emobj.SetOfCTF(filename=self._dbName,
                                     prefix=self._dbPrefix)

        count = 0
        for ctf in modifiedSet:
            if self._itemSelected(ctf):
                mic = ctf.getMicrograph()
                outputMics.append(mic)
                outputCtfs.append(ctf)
                count += 1

        # Register outputs
        outputCtfs.setMicrographs(outputMics)
        # NOTE: I've split the define output in 2 steps.
        # It seems with python3 outputCTF was processed first and needs mics to be saved first.
        self._defineOutputs(outputMicrographs=outputMics)
        self._defineOutputs(outputCTF=outputCtfs)
        self._defineTransformRelation(setOfMics, outputMics)
        self._defineCtfRelation(outputMics, outputCtfs)
        msg = 'From input %s of size %s created output ' % (inputCTFs.getClassName(),
                                                            inputCTFs.getSize())
        msg += 'SetOfMicrographs and SetOfCTF of size %d' % count
        self.summaryVar.set(msg)

        return outputMics, outputCtfs

    def _createRepresentativesFromClasses(self, inputClasses, outputClassName):
        """ Create a new set of images joining all images
        assigned to each class.
        """
        inputImages = inputClasses.getImages()

        if isinstance(outputClassName, str):
            createFunc = getattr(self, '_create' + outputClassName)
            output = createFunc()
        else:
            output = outputClassName.create(self.getPath())

        modifiedSet = inputClasses.getClass()(filename=self._dbName, prefix=self._dbPrefix)
        self.info("Creating REPRESENTATIVES of images from classes, "
                  "sqlite file: %s" % self._dbName)

        count = 0

        output.copyInfo(inputImages)
        output.setSamplingRate(None)
        # For now this is to avoid having a wrong alignment.
        # THis is because is getting the alignment info from the input images and this does not have to match.
        # This created an error when scaling averages #903
        output.setAlignment(ALIGN_NONE)
        for cls in modifiedSet:
            if self._itemSelected(cls):
                img = cls.getRepresentative()
                if not output.getSamplingRate():
                    output.setSamplingRate(cls.getSamplingRate()
                                           if img.getSamplingRate() is None
                                           else img.getSamplingRate())
                img.copyObjId(cls)
                output.append(img)
                count += 1
        # Register outputs
        self._defineOutput('Representatives', output)
        if inputClasses.hasObjId():
            self._defineSourceRelation(inputClasses, output)
        else:
            self._defineSourceRelation(inputImages, output)

        selectmsg = 'we selected %s items' % count if count > 1 else 'was selected 1 item'
        msg = 'From input %s of size %s %s to create output %s' % (inputClasses.getClassName(),
                                                                   inputClasses.getSize(),
                                                                   selectmsg,
                                                                   output.getClassName())
        self.summaryVar.set(msg)
        return output

    @staticmethod
    def _copyInfoAndSetAlignment(inputClasses, output):
        """ This method is used when creating subset of images from classes.
        We need to copy the information from the original classes images
        and also set the proper alignment contained in the classes.
        """
        inputImages = inputClasses.getImages()
        # Copy all info form the original 'classified' images
        output.copyInfo(inputImages)
        # Take the alignment of the first class
        cls = inputClasses.getFirstItem()
        output.setAlignment(cls.getAlignment())

    def _createImagesFromClasses(self, inputClasses):
        """ Create a new set of images joining all images
        assigned to each class.
        """
        inputImages = inputClasses.getImages()
        className = inputImages.getClassName()
        try:
            createFunc = getattr(self, '_create' + className)
            output = createFunc()
        except Exception as e:
            output = inputImages.createCopy(self._getPath())

        modifiedSet = inputClasses.getClass()(filename=self._dbName, prefix=self._dbPrefix)
        self.info("Creating subset of images from classes, sqlite file: %s" % self._dbName)
        self._copyInfoAndSetAlignment(inputClasses, output)
        output.appendFromClasses(modifiedSet, filterClassFunc=self._itemSelected)
        # Register outputs
        self._defineOutput(className, output)
        if inputClasses.hasObjId():
            self._defineSourceRelation(inputClasses, output)
        self._defineTransformRelation(inputImages, output)
        count = len([cls for cls in modifiedSet if cls.isEnabled()])
        selectmsg = 'we selected %s items' % count if count > 1 else 'was selected 1 item'
        msg = 'From input %s of size %s %s to create output %s of size %s' % (inputClasses.getClassName(),
                                                                              inputClasses.getSize(),
                                                                              selectmsg,
                                                                              output.getClassName(),
                                                                              output.getSize())
        self.summaryVar.set(msg)
        return output

    def _createClassesFromClasses(self, inputClasses):
        """ Create a new set of images joining all images
        assigned to each class.
        """
        # inputImages = inputClasses.getImages()
        className = inputClasses.getClassName()
        # createFunc = getattr(self, '_create' + className)

        try:
            createFunc = getattr(self, '_create' + className)
            output = createFunc()
        except Exception as e:
            output = inputClasses.getClass().create(self._getPath())

        modifiedSet = inputClasses.getClass()(filename=self._dbName, prefix=self._dbPrefix)
        self.info("Creating subset of classes from classes, sqlite file: %s" % self._dbName)
        output.copyInfo(inputClasses)
        output.appendFromClasses(modifiedSet,filterClassFunc=self._itemSelected)
        # Register outputs
        self._defineOutput(className, output)
        if inputClasses.hasObjId():
            self._defineTransformRelation(inputClasses, output)
        else:
            self._defineSourceRelation(inputClasses.getImages(), output)
        count = len([cls for cls in modifiedSet if cls.isEnabled()])
        selectmsg = 'we selected %s items' % count if count > 1 else 'was selected 1 item'
        msg = 'From input %s of size %s %s to create output %s' % (inputClasses.getClassName(),
                                                                   inputClasses.getSize(),
                                                                   selectmsg,
                                                                   output.getClassName())
        self.summaryVar.set(msg)
        return output

    def _createSubSetFromMicrographsTiltPair(self, micrographsTiltPair):
        """ Create a subset of Micrographs Tilt Pair. """
        output = MicrographsTiltPair(filename=self._getPath('micrographs_pairs.sqlite'))
        modifiedSet = MicrographsTiltPair(filename=self._dbName,
                                          prefix=self._dbPrefix)
        inputU = micrographsTiltPair.getUntilted()
        inputT = micrographsTiltPair.getTilted()
        outputU = emobj.SetOfMicrographs(filename=self._getPath('mics_untilted.sqlite'))
        outputT = emobj.SetOfMicrographs(filename=self._getPath('mics_tilted.sqlite'))
        outputU.copyInfo(inputU)
        outputT.copyInfo(inputT)

        # noinspection DuplicatedCode
        for micPair, u, t in izip(modifiedSet, inputU, inputT):
            if micPair.isEnabled():
                output.append(micPair)
                outputU.append(u)
                outputT.append(t)
        output.setUntilted(outputU)
        output.setTilted(outputT)
        # Register outputs
        outputDict = {'outputMicrographsTiltPair': output}
        self._defineOutputs(**outputDict)
        self._defineTransformRelation(micrographsTiltPair, output)
        return output

    def _createSubSetFromAtomStructs(self, setOfPDBs):
        """ Create a subset of SetOfAtomStruct. """
        output = emobj.SetOfAtomStructs(filename=self._getPath('atomstructs.sqlite'))
        modifiedSet = emobj.SetOfAtomStructs(filename=self._dbName, prefix=self._dbPrefix)

        for pdb in modifiedSet:
            if pdb.isEnabled():
                output.append(pdb)

        # Register outputs
        outputDict = {'outputAtomStructs': output}
        self._defineOutputs(**outputDict)
        self._defineTransformRelation(setOfPDBs, output)
        return output

    def _createSubSetFromParticlesTiltPair(self, particlesTiltPair):
        """ Create a subset of Particles Tilt Pair. """
        output = ParticlesTiltPair(filename=self._getPath('particles_pairs.sqlite'))

        inputU = particlesTiltPair.getUntilted()
        inputT = particlesTiltPair.getTilted()
        outputU = emobj.SetOfParticles(filename=self._getPath('particles_untilted.sqlite'))
        outputT = emobj.SetOfParticles(filename=self._getPath('particles_tilted.sqlite'))
        outputU.copyInfo(inputU)
        outputT.copyInfo(inputT)

        modifiedSet = ParticlesTiltPair(filename=self._dbName,
                                        prefix=self._dbPrefix)

        # noinspection DuplicatedCode
        for pair, u, t in izip(modifiedSet, inputU, inputT):
            if pair.isEnabled():
                output.append(pair)
                outputU.append(u)
                outputT.append(t)
        # Register outputs
        output.setUntilted(outputU)
        output.setTilted(outputT)
        # Link output to the same coordinates pairs than input
        output.setCoordsPair(particlesTiltPair.getCoordsPair())

        outputDict = {'outputParticlesTiltPair': output}
        self._defineOutputs(**outputDict)
        self._defineTransformRelation(particlesTiltPair, output)
        return output

    # noinspection PyAttributeOutsideInit
    def createSetObject(self):
        """ Moves the sqlite with the enable/disable status to its own
        path to keep it and names it subset.sqlite"""

        _dbName, self._dbPrefix = self.sqliteFile.get().split(',')

        if self.usingShowJ():

            self._dbName = self._getPath('subset.sqlite')
            os.rename(_dbName, self._dbName)
        else:
            self._dbName = self.inputObject.get().getFileName()
            self._selectionTxt = _dbName

        # Prefix: used to create subsets from Particles from a class (specific table in a set of classes)
        if self._dbPrefix.endswith('_'):
            self._dbPrefix = self._dbPrefix[:-1]

        from pwem.utils import loadSetFromDb

        # Ignoring self._dbPrefix here, since we want to load
        # the top-level set in the sqlite file
        setObj = loadSetFromDb(self._dbName)
        return setObj

    def _summary(self):
        summary = []
        msg = self.summaryVar.get()
        if msg is None:
            msg = self.getDefaultSummary()
        summary.append(msg)
        return summary

    def getDefaultSummary(self):
        inputStr = ''
        inputObj = self.inputObject.get()
        if inputObj is not None:
            inputStr += inputObj.getClassName()
            if isinstance(inputObj, emobj.EMSet):
                inputStr += ' of size %s' % inputObj.getSize()
        output = ''
        for _, attr in self.iterOutputAttributes():
            output += attr.getClassName()
            if isinstance(attr, emobj.EMSet):
                output += ' of size %s' % attr.getSize()

        msg = 'From input %s created output %s ' % (inputStr, output)

        return msg

    def _methods(self):
        return self._summary()

    def _defineOutput(self, className, output):
        outputDict = {'output' + className.replace('SetOf', ''): output}
        self._defineOutputs(**outputDict)


class ProtCreateMask(BatchProtocol):

    def _defineParams(self, form):
        form.addHidden('inputObj', PointerParam, pointerClass='EMObject')
        form.addHidden('maskFile', StringParam)

    def _insertAllSteps(self):
        self._insertFunctionStep('createMaskStep')

    def createMaskStep(self):
        inputObj = self.inputObj.get()
        maskSrc = self.maskFile.get()
        basename = os.path.basename(maskSrc)
        maskDst = self._getPath(basename)
        moveFile(maskSrc, maskDst)
        samplingRate = None
        if hasattr(inputObj, "getSamplingRate"):
            samplingRate = inputObj.getSamplingRate()
        else:
            for key, attr in inputObj.iterInputAttributes():
                if hasattr(attr.get(), "getSamplingRate"):
                    samplingRate = attr.get().getSamplingRate()
        if not samplingRate:
            raise Exception("sampling rate required")

        mask = emobj.Mask()
        mask.setFileName(maskDst)
        mask.setSamplingRate(samplingRate)
        self._defineOutputs(outputMask=mask)
        self._defineSourceRelation(self.inputObj, self.outputMask)

    def _summary(self):
        summary = list()
        summary.append('From input %s created mask %s'
                       % (self.getObjectTag("inputObj"),
                          self.getObjectTag("outputMask")))
        return summary

    def _methods(self):
        return self._summary()


class ProtCreateFSC(BatchProtocol):

    def _defineParams(self, form):
        form.addHidden('inputObj', PointerParam,
                       pointerClass='EMObject')
        form.addHidden('fscValues', StringParam,
                       help='String representation of the list with FSC values')
        form.addHidden('fscLabels', StringParam,
                       help='String with fsc labels')

    def setInputObj(self, obj):
        self.inputObj.set(obj)

    def setInputFscList(self, fscList):
        fscStr = ''
        fscLabel = ""
        numberFsc = 0
        for fsc in fscList:
            if numberFsc == 0:
                numberFsc += 1
            else:
                fscStr += "|"
                fscLabel += "|"
            fscStr += dumps(fsc.getData())
            fscLabel += dumps(fsc.getObjLabel())
        self.fscValues.set(fscStr)
        self.fscLabels.set(fscLabel)

    def _insertAllSteps(self):
        self._insertFunctionStep('createOutputStep')

    def createOutputStep(self):
        fscSet = self._createSetOfFSCs()
        fscSet.setObjLabel("setOfFSCs")
        dataStringList = self.fscValues.get().split("|")
        labelStringList = self.fscLabels.get().split("|")
        for fsc, label in zip(dataStringList, labelStringList):
            _fsc = emobj.FSC(objLabel=loads(label))
            freq, value = loads(fsc)
            _fsc.setData(freq, value)
            fscSet.append(_fsc)
        self._defineOutputs(outputFSCs=fscSet)

    def _summary(self):
        summary = list()
        summary.append('From input %s created FSC %s'
                       % (self.getObjectTag("inputObj"),
                          self.getObjectTag("outputFSCs")))
        return summary

    def _methods(self):
        return self._summary()




# ***************************************************************************
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# ***************************************************************************/
import os

import pyworkflow.tests as pwtests

import pwem.protocols as emprot
import pwem.constants as emcts
import pwem.convert as emconv
from pwem.convert import Ccp4Header
from pyworkflow.object import Pointer


class TestImportBase(pwtests.BaseTest):
    @classmethod
    def setUpClass(cls):
        pwtests.setupTestProject(cls)
        cls.dsModBuild = pwtests.DataSet.getDataSet('model_building_tutorial')


class TestImportAtomStruct(TestImportBase):

    def test_import_atomStruct(self):
        """ 1) Import single atomstruct form file
        """
        args = {'inputPdbData': emprot.ProtImportPdb.IMPORT_FROM_FILES,
                'pdbFile': self.dsModBuild.getFile('PDBx_mmCIF/1ake_mut1.pdb'),
                }

        prot1 = self.newProtocol(emprot.ProtImportPdb, **args)
        prot1.setObjLabel('import atomstruct from file')
        self.launchProtocol(prot1)
        self.assertIsNotNone(prot1.outputPdb)


        # """ 2) Import single atomstruct from pdb id
        # """
        args = {'inputPdbData': emprot.ProtImportPdb.IMPORT_FROM_ID,
                'pdbId': '1ake',
                }

        # Id's should be set increasing from 1 if ### is not in the
        # pattern
        prot2 = self.newProtocol(emprot.ProtImportPdb, **args)
        prot2.setObjLabel('import atomstruct from pdb id')
        self.launchProtocol(prot2)
        self.assertIsNotNone(prot2.outputPdb)


    def test_import_setOfAtomStruct(self):
        # """ 1) Import set of atomstruct from files
        # """
        args = {'inputPdbData': emprot.ProtImportSetOfAtomStructs.IMPORT_FROM_FILES,
                'filesPath': self.dsModBuild.getFile('PDBx_mmCIF'),
                'filesPattern': self.dsModBuild.getFile('PDBx_mmCIF/1ake_*.pdb')
                }
        prot1 = self.newProtocol(emprot.ProtImportSetOfAtomStructs, **args)
        prot1.setObjLabel('import 1akes structures from files\n')
        self.launchProtocol(prot1)
        self.assertIsNotNone(getattr(prot1, prot1._OUTNAME))

        # """ 2) Import set of atomstruct form pdb id
        # """
        args = {'inputPdbData': emprot.ProtImportSetOfAtomStructs.IMPORT_FROM_ID,
                'pdbIds': '1ake,5ni1',
                }
        prot2 = self.newProtocol(emprot.ProtImportSetOfAtomStructs, **args)
        prot2.setObjLabel('import 1ake and 5ni1 structures from pdb ids\n')
        self.launchProtocol(prot2)
        self.assertIsNotNone(getattr(prot2, prot2._OUTNAME))


# **************************************************************************
# *
# * Authors:     J.M. De la Rosa Trevin (jmdelarosa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

from os.path import abspath

import pyworkflow.utils as pwutils
import pyworkflow.protocol.constants as pwcts
import pyworkflow.protocol.params as params

from pwem import Domain
import pwem.constants as emcts

from .images import ProtImportImages


class ProtImportParticles(ProtImportImages):
    """Protocol to import a set of particles to the project"""
    _label = 'import particles'
    _outputClassName = 'SetOfParticles'

    IMPORT_FROM_EMX = 1
    IMPORT_FROM_XMIPP3 = 2
    IMPORT_FROM_RELION = 3
    IMPORT_FROM_SCIPION = 4
    IMPORT_FROM_FREALIGN = 5
    IMPORT_FROM_EMAN = 6
    IMPORT_FROM_CRYOSPARC = 7

    importFormats = ['emx', 'xmipp3', 'relion', 'scipion', 'frealign', 'eman', 'cryosparc']
    importExts = ['emx', 'xmd', 'star', 'sqlite', 'par', 'lst', 'cs']
    alignTypeList = [emcts.ALIGN_2D, emcts.ALIGN_3D, emcts.ALIGN_PROJ,
                     emcts.ALIGN_NONE]

    def _getImportChoices(self):
        """ Return a list of possible choices
        from which the import can be done.
        (usually packages formats such as: xmipp3, eman2, relion...etc.)
        """
        choices = ProtImportImages._getImportChoices(self)
        # Do not change the order of this list since
        # it is related to the constants defined
        return choices + self.importFormats
    
    def _defineImportParams(self, form):
        """
        Import files from: emx, xmipp3, relion, scipion, cryosparc formats.
        """
        param = form.getParam('importFrom')
        # Customize the help of this parameter with specific information
        # of the import particles
        param.help.set("You can import particles directly from the binary "
                       "files, or import from other packages formats. \n"
                       "Currently, we can import from: %s \n"
                       "Following are the expected import files for each one:\n"
                       "*emx*: particles.emx\n"
                       "*xmipp3*: images.xmd\n"
                       "*relion*: itXX_data.star\n"
                       "*scipion*: particles.sqlite\n"
                       "*eman*: particleSet.lst\n"
                       "*cryosparc*: particles.cs\n"
                       "" % ', '.join(self.importFormats))

        form.addParam('emxFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_EMX,
                      label='Input EMX file',
                      help="Select the EMX file containing particles "
                           "information.\n See more about \n"
                           "[[http://i2pc.cnb.csic.es/emx][EMX format]]")

        form.addParam('alignType', params.EnumParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_EMX,
                      default=0,
                      choices=self.alignTypeList,
                      label='Alignment Type',
                      help="Is this a 2D alignment, a 3D alignment or a set of projections")
#
        form.addParam('mdFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_XMIPP3,
                      label='Particles metadata file',
                      help="Select the particles Xmipp metadata file.\n"
                           "It is usually a images.xmd file result\n"
                           "from Xmipp protocols execution.")
        
        form.addParam('starFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_RELION,
                      label='Star file',
                      help="Select a *_data.star file from a\n"
                           "previous Relion execution."
                           "To detect if the input particles contains alignment "
                           "information, it is required to have the "
                           "optimiser.star file corresponding to the data.star")
        
        form.addParam('ignoreIdColumn', params.BooleanParam, default=False,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_RELION,
                      label='Ignore ID column?',
                      help="Set this option to True to regenerate \n"
                           "the id's of the particles. By default \n"
                           "it is read from metadata file.        \n"
                           "This option can be useful when merging\n"
                           "different metadatas and id's are not  \n"
                           "longer unique.")
        
        form.addParam('sqliteFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_SCIPION,
                      label='Particles sqlite file',
                      help="Select the particles sqlite file.\n")

        form.addParam('frealignLabel', params.LabelParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_FREALIGN,
                      label='For Frealign you need to import both stack and .par files.')  
        form.addParam('stackFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_FREALIGN,
                      label='Stack file',
                      help="Select an stack file with the particles.")          
        form.addParam('parFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_FREALIGN,
                      label='Param file',
                      help="Select a Frealign .par file with the refinement information.")
        form.addParam('lstFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_EMAN,
                      label='Lst file',
                      help='Select a *.lst set file from EMAN2 project.')

        form.addParam('csFile', params.FileParam,
                      condition='(importFrom == %d)' % self.IMPORT_FROM_CRYOSPARC,
                      label='cs file',
                      help="Select a .cs file.\n"
                           "It is usually a .cs file result from cryoSPARC job "
                           "execution.")
        
    def _defineAcquisitionParams(self, form):
        group = ProtImportImages._defineAcquisitionParams(self, form)
        group.addParam('samplingRate', params.FloatParam,
                       label=pwutils.Message.LABEL_SAMP_RATE)

    def _insertAllSteps(self):
        importFrom = self.importFrom.get()
        ci = self.getImportClass()
        
        if ci is None:
            ProtImportImages._insertAllSteps(self)
        else:
            self._insertFunctionStep('importParticlesStep', importFrom,
                                     self.importFilePath)
            
    def getImportClass(self):
        """ Return the class in charge of importing the files. """
        if self.importFrom == self.IMPORT_FROM_EMX:
            EmxImport = Domain.importFromPlugin('emxlib.convert', 'EmxImport',
                                                errorMsg='Emx is needed to import .emx files',
                                                doRaise=True)
            self.importFilePath = abspath(self.emxFile.get('').strip())
            return EmxImport(self, self.importFilePath,
                             self.alignTypeList[self.alignType.get()])

        elif self.importFrom == self.IMPORT_FROM_XMIPP3:
            XmippImport = Domain.importFromPlugin('xmipp3.convert', 'XmippImport',
                                                  'Xmipp is needed to import .xmd files',
                                                  doRaise=True)
            self.importFilePath = self.mdFile.get('').strip()
            return XmippImport(self, self.mdFile.get())

        elif self.importFrom == self.IMPORT_FROM_RELION:
            RelionImport = Domain.importFromPlugin('relion.convert', 'RelionImport',
                                                   errorMsg='Relion is needed to import .star files',
                                                   doRaise=True)
            self.importFilePath = self.starFile.get('').strip()
            return RelionImport(self, self.starFile.get())
        elif self.importFrom == self.IMPORT_FROM_SCIPION:
            from .dataimport import ScipionImport
            self.importFilePath = self.sqliteFile.get('').strip()
            return ScipionImport(self, self.importFilePath)    
        elif self.importFrom == self.IMPORT_FROM_FREALIGN:
            self.importFilePath = self.parFile.get('').strip()
            GrigorieffLabImportParticles = Domain.importFromPlugin(
                'cistem.convert', 'GrigorieffLabImportParticles',
                errorMsg='Cistem is needed to import .stk files',
                doRaise=True)
            return GrigorieffLabImportParticles(self, self.parFile.get(),
                                                self.stackFile.get())
        elif self.importFrom == self.IMPORT_FROM_EMAN:
            self.importFilePath = self.lstFile.get('').strip()
            EmanImport = Domain.importFromPlugin('eman2.convert', 'EmanImport',
                                                 doRaise=True)
            return EmanImport(self, self.lstFile.get())

        elif self.importFrom == self.IMPORT_FROM_CRYOSPARC:
            cryoSPARCImport = Domain.importFromPlugin('cryosparc2.convert', 'cryoSPARCImport',
                                                      'cryoSPARC is needed to import .cs files',
                                                      doRaise=True)
            self.importFilePath = self.csFile.get('').strip()
            return cryoSPARCImport(self, self.csFile.get())
        else:
            self.importFilePath = ''
            return None 
                    
    def setSamplingRate(self, imgSet):
        imgSet.setSamplingRate(self.samplingRate.get())
    
    def importParticlesStep(self, importFrom, *args):
        ci = self.getImportClass()
        ci.importParticles()
        
        summary = "Import from *%s* file:\n" % self.getEnumText('importFrom')
        summary += self.importFilePath + '\n'
        
        if self.hasAttribute('outputParticles'):
            particles = self.outputParticles
            summary += ' Particles: *%d* ' % particles.getSize()
            summary += ('(ctf=%s, alignment=%s, phaseFlip=%s)\n'
                        % (particles.hasCTF(), particles.getAlignment(),
                           particles.isPhaseFlipped()))

        # EMX files can contain only Coordinates information
        if self.hasAttribute('outputCoordinates'):
            summary += '   Coordinates: *%d* \n' % (self.outputCoordinates.getSize())

        # EMX files can contain only Coordinates information
        if self.hasAttribute('outputMicrographs'):
            summary += '   Micrographs: *%d* \n' % (self.outputMicrographs.getSize())
        
        if self.copyFiles:
            summary += '\n_WARNING_: Binary files copied into project (extra disk space)'
            
        self.summaryVar.set(summary)

    def _validateFileExtension(self):
        """ Simple check about the expected file extension. """
        # Since 'from files' is index 0, we need to subtract 1 to
        # get the proper index
        i = self.importFrom.get() - 1
        ext = self.importExts[i]
        if not self.importFilePath.endswith(ext):
            return ["Expected *%s* file extension for importing from *%s*" %
                    (ext, self.importFormats[i])]
        else:
            return []

    def _validate(self):
        ci = self.getImportClass()
        if ci is None:
            return ProtImportImages._validate(self)
        else:
            errors = self._validateFileExtension()
            if errors:
                return errors
            else:
                return ci.validateParticles()
    
    def _summary(self):
        if self.importFrom == self.IMPORT_FROM_FILES:
            return ProtImportImages._summary(self)
        else:
            return [self.summaryVar.get('')]


class ProtImportAverages(ProtImportParticles):
    """Protocol to import a set of averages to the project"""
    _label = 'import averages'
    _outputClassName = 'SetOfAverages'    

    def _getImportChoices(self):
        """ Return a list of possible choices
        from which the import can be done.
        (usually packages formas such as: xmipp3, eman2, relion...etc.
        """
        choices = ProtImportImages._getImportChoices(self)
        return choices
            
    def _defineAcquisitionParams(self, form):
        form.addParam('samplingRate', params.FloatParam, default=1.,
                      label=pwutils.Message.LABEL_SAMP_RATE)
        group = ProtImportImages._defineAcquisitionParams(self, form)
        group.expertLevel.set(pwcts.LEVEL_ADVANCED)



# **************************************************************************
# *
# * Authors:     Roberto Marabini (roberto@cnb.csic.es)
# *              Marta Martinez (mmmtnez@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 3 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************

import os

import pyworkflow.utils as pwutils
import pyworkflow.viewer as pwviewer
import pwem.constants as emcts
import pwem.emlib.metadata as md
import pwem.objects as emobj
from pwem import emlib
from pwem import Config as emConfig
from pwem.objects import SetOfAtomStructs

chimeraPdbTemplateFileName = "Atom_struct__%06d.cif"
chimeraMapTemplateFileName = "Map__%06d.mrc"
chimeraScriptFileName = "chimeraScript.cxc"
chimeraConfigFileName = "chimera.ini"
sessionFile = "SESSION.cxs"

symMapperScipionchimera = {}
symMapperScipionchimera[emcts.SYM_CYCLIC] = "Cn"
symMapperScipionchimera[emcts.SYM_DIHEDRAL] = "Dn"
symMapperScipionchimera[emcts.SYM_TETRAHEDRAL] = "T"
symMapperScipionchimera[emcts.SYM_OCTAHEDRAL] = "O"
symMapperScipionchimera[emcts.SYM_I222] = "222"
symMapperScipionchimera[emcts.SYM_I222r] = "222r"
symMapperScipionchimera[emcts.SYM_In25] = "n25"
symMapperScipionchimera[emcts.SYM_In25r] = "n25r"
symMapperScipionchimera[emcts.SYM_I2n3] = "2n3"
symMapperScipionchimera[emcts.SYM_I2n3r] = "2n3r"
symMapperScipionchimera[emcts.SYM_I2n5] = "2n5"
symMapperScipionchimera[emcts.SYM_I2n5r] = "2n5r"


class Chimera:
    """ Helper class to execute chimera and handle its environment. """

    # Map symmetries from Scipion convention to Chimera convention
    _symmetryMap = {
        emcts.SYM_CYCLIC: 'Cn',
        emcts.SYM_DIHEDRAL: 'Dn',
        emcts.SYM_TETRAHEDRAL: 'T',
        emcts.SYM_OCTAHEDRAL: 'O',
        emcts.SYM_I222: '222',
        emcts.SYM_I222r: '222r',
        emcts.SYM_In25: 'n25',
        emcts.SYM_In25r: 'n25r',
        emcts.SYM_I2n3: '2n3',
        emcts.SYM_I2n3r: '2n3r',
        emcts.SYM_I2n5: '2n5',
        emcts.SYM_I2n5r: '2n5r'
    }

    @classmethod
    def getSymmetry(cls, scipionSym):
        """ Return the equivalent Chimera symmetry from Scipion one. """
        return cls._symmetryMap[scipionSym]

    @classmethod
    def getHome(cls):
        """ Returns chimera home, trying first to take it from chimera plugin. If it fails it will return, the default value in the config"""
        with pwutils.weakImport("chimera"):
            from chimera import Plugin as chimeraPlugin
            return chimeraPlugin.getHome()

        return os.path.join(emConfig.EM_ROOT, 'chimerax-1.2.5')

    @classmethod
    def getEnviron(cls):
        """ Return the proper environ to launch chimera.
        CHIMERA_HOME variable is read from the ~/.config/scipion.conf file.
        """
        environ = pwutils.Environ(os.environ)
        environ.set('PATH', os.path.join(cls.getHome(), 'bin'),
                    position=pwutils.Environ.BEGIN)

        if "REMOTE_MESA_LIB" in os.environ:
            environ.set('LD_LIBRARY_PATH', os.environ['REMOTE_MESA_LIB'],
                        position=pwutils.Environ.BEGIN)
        return environ

    @classmethod
    def getProgram(cls, progName="ChimeraX"):
        """ Return the program binary that will be used. """
        home = cls.getHome()
        if home is None:
            return None
        return os.path.join(home, 'bin', os.path.basename(progName))

    @classmethod
    def runProgram(cls, program=None, args="", cwd=None):
        """ Internal shortcut function to launch chimera program. """
        prog = program or cls.getProgram()
        pwutils.runJob(None, prog, args, env=cls.getEnviron(),
                       cwd=cwd)

    @classmethod
    def createCoordinateAxisFile(cls, dim, bildFileName="/tmp/axis.bild",
                                 sampling=1, r1=0.1):
        """ Create a coordinate system, Along each dimension
        we place a small sphere in the negative axis. In this way
        chimera shows the system of coordinates origin in the
        window center"""

        ff = open(bildFileName, "w")
        arrowDict = {}
        arrowDict["x"] = arrowDict["y"] = arrowDict["z"] = \
            sampling * dim * 1. / 2.
        arrowDict["r1"] = r1 * dim / 50.
        arrowDict["r2"] = 2 * arrowDict["r1"]
        arrowDict["rho"] = 0.75  # axis thickness

        ff.write(".color red\n" # red
                 ".arrow 0 0 0 %(x)f 0 0 %(r1)f %(r2)f %(rho)f\n"
                 ".color 0 0 0\n.sphere -%(x)f 0 0 0.00001\n"
                 ".color yellow\n" # yellow
                 ".arrow 0 0 0 0 %(y)f 0 %(r1)f %(r2)f %(rho)f\n"
                 ".color 0 0 0\n.sphere 0 -%(y)f 0 0.00001\n"
                 ".color blue\n"
                 ".arrow 0 0 0 0 0 %(z)f %(r1)f %(r2)f %(rho)f\n"
                 ".color 0 0 0\n.sphere 0 0 -%(z)f 0.00001\n" %
                 arrowDict)
        ff.close()


class ChimeraAngDist(pwviewer.CommandView):

    def __init__(self, volFile, tmpFilesPath, **kwargs):
        self.kwargs = kwargs
        self.volfile = os.path.abspath(os.path.join(volFile))
        self.cmdFile = os.path.join(tmpFilesPath,
                                    "chimera_angular_distribution.cxc")
        self.axis = os.path.abspath(os.path.join(tmpFilesPath, "axis.bild"))
        self.spheres = os.path.abspath(os.path.join(tmpFilesPath,
                                                    "spheres.bild"))

        self.angularDistributionList = []
        self.angularDistFile = kwargs.get('angularDistFile', None)
        if self.angularDistFile:
            fileName = self.angularDistFile
            if '@' in self.angularDistFile:
                fileName = self.angularDistFile.split("@")[1]
            if not (os.path.exists(fileName)):  # check blockname:
                raise Exception("Path %s does not exists" %
                                self.angularDistFile)

        self.volOrigin = kwargs.get('volOrigin', None)
        self.spheresColor = kwargs.get('spheresColor', 'orange')
        spheresDistance = kwargs.get('spheresDistance', -1)
        spheresMaxRadius = kwargs.get('spheresMaxRadius', None)
        self.initVolumeData(self.volfile)
        self.spheresDistance = (float(spheresDistance) if spheresDistance != -1
                                else (0.75 * self.voxelSize * max(self.xdim,
                                                                  self.ydim,
                                                                  self.zdim)) / 2)
        self.spheresMaxRadius = (float(spheresMaxRadius) if spheresMaxRadius
                                 else 0.02 * self.spheresDistance)
        self.readAngularDistFile()
        format = kwargs.get('format', None)
        self._createChimeraScript(self.cmdFile, format)
        program = Chimera.getProgram()
        pwviewer.CommandView.__init__(self, '%s "%s" &' % (program,
                                                           self.cmdFile),
                                      env=Chimera.getEnviron(), **kwargs)

    def _createChimeraScript(self, scriptFile, format):

        Chimera.createCoordinateAxisFile(self.xdim,
                                         bildFileName=self.axis,
                                         sampling=self.voxelSize)
        self.createAngularDistributionFile(bildFileName=self.spheres)
        fhCmd = open(scriptFile, 'w')
        fhCmd.write("open %s\n" % self.axis)
        fhCmd.write("open %s\n" % self.spheres)
        fhCmd.write("cofr 0,0,0\n")
        if format is not None:
            fhCmd.write("open %s format %s\n" % (self.volfile.replace(":mrc", ""),
                                                 format))
        else:
            fhCmd.write("open %s\n" % self.volfile.replace(":mrc", ""))
        fhCmd.write("volume #3 voxelSize %s\n" % self.voxelSize)
        x, y, z = self.volOrigin
        fhCmd.write("volume #3 origin %s,%s,%s\n" % (x, y, z))
        fhCmd.write("view\n")

        fhCmd.close()

    def createAngularDistributionFile(self, bildFileName="/tmp/spheres.bild"):
        ff = open(bildFileName, "w")

        for angulardist in self.angularDistributionList:
            ff.write("%s\n" % angulardist)

        ff.close()

    def initVolumeData(self, volfile):
        if volfile.endswith('.mrc'):
            volfile += ':mrc'

        if volfile is None:
            raise ValueError(volfile)
        if '@' in volfile:
            [index, file] = volfile.split('@')
        else:
            file = volfile
        if ':' in file:
            file = file[0: file.rfind(':')]
        if not os.path.exists(file):
            raise Exception("File %s does not exists" % file)

        self.voxelSize = self.kwargs.get('voxelSize', 1.0)
        self.image = emlib.Image(self.volfile)
        self.image.convert2DataType(md.DT_DOUBLE)
        self.xdim, self.ydim, self.zdim, self.n = self.image.getDimensions()
        self.vol = self.image.getData()

    def readAngularDistFile(self):
        angleRotLabel = md.MDL_ANGLE_ROT
        angleTiltLabel = md.MDL_ANGLE_TILT
        anglePsiLabel = md.MDL_ANGLE_PSI
        mdAngDist = md.MetaData(self.angularDistFile)
        if not mdAngDist.containsLabel(md.MDL_ANGLE_PSI):
            anglePsiLabel = None
            if mdAngDist.containsLabel(md.RLN_ORIENT_PSI):
                angleRotLabel = md.RLN_ORIENT_ROT
                angleTiltLabel = md.RLN_ORIENT_TILT
                anglePsiLabel = md.RLN_ORIENT_PSI

        if not mdAngDist.containsLabel(md.MDL_WEIGHT):
            mdAngDist.fillConstant(md.MDL_WEIGHT, 1.)

        maxweight = mdAngDist.aggregateSingle(md.AGGR_MAX, md.MDL_WEIGHT)
        minweight = mdAngDist.aggregateSingle(md.AGGR_MIN, md.MDL_WEIGHT)
        interval = maxweight - minweight
        x2 = self.xdim / 2
        y2 = self.ydim / 2
        z2 = self.zdim / 2
        self.angularDistributionList.append('.color %s\n' % self.spheresColor)
        for id in mdAngDist:
            rot = mdAngDist.getValue(angleRotLabel, id)
            tilt = mdAngDist.getValue(angleTiltLabel, id)
            psi = mdAngDist.getValue(anglePsiLabel, id) if anglePsiLabel else 0
            weight = mdAngDist.getValue(md.MDL_WEIGHT, id)
            # Avoid zero division
            weight = 0 if interval == 0 else (weight - minweight) / interval
            weight = weight + 0.5  # add 0.5 to avoid zero weight
            x, y, z = emlib.Euler_direction(rot, tilt, psi)
            radius = weight * self.spheresMaxRadius

            x = x * self.spheresDistance  # + x2
            y = y * self.spheresDistance  # + y2
            z = z * self.spheresDistance  # + z2
            command = ('.sphere %.1f %.1f %.1f %.1f' %
                       (x, y, z, radius))
            self.angularDistributionList.append(command)


class ChimeraView(pwviewer.CommandView):
    """ View for calling an external command. """

    def getProgram(self):
        return Chimera.getProgram()

    def getEnviron(self):
        return Chimera.getEnviron()

    def __init__(self, inputFiles, **kwargs):

        if isinstance(inputFiles,str):
            inputFiles = [inputFiles]

        program = self.getProgram()

        # Join files
        filesStr = '" "'.join(inputFiles)
        filesStr = filesStr.replace(":mrc", "")
        pwviewer.CommandView.__init__(self, '%s "%s" &' % (program, filesStr),
                                      env=self.getEnviron(), **kwargs)

class ChimeraViewer(pwviewer.Viewer):
    """ Wrapper to visualize PDB object with Chimera. """
    _environments = [pwviewer.DESKTOP_TKINTER]
    _targets = [emobj.AtomStruct, emobj.PdbFile, emobj.SetOfAtomStructs,
                emobj.Volume, emobj.SetOfVolumes, emobj.SetOfClasses3D]

    _name = "ChimeraX"

    def __init__(self, **kwargs):
        pwviewer.Viewer.__init__(self, **kwargs)

    def _visualize(self, obj, **kwargs):
        cls = type(obj)
        if issubclass(cls, emobj.AtomStruct):
            objSet = SetOfAtomStructs.create(outputPath='/tmp', suffix=self.protocol.getObjId())
            objSet.append(obj)
            if hasattr(obj, '_chimeraScript'):
                objSet.copyAttributes(obj, '_chimeraScript')

            obj, cls = objSet, type(objSet)

        if issubclass(cls, emobj.SetOfAtomStructs):
            if hasattr(obj, '_chimeraScript'):
                fn = obj._chimeraScript.get()
                view = ChimeraView(fn)
                return [view]
            else:
                fnCmd = self.protocol._getExtraPath("chimera_output.cxc")

                f = open(fnCmd, 'w')
                f.write('cd %s\n' % os.getcwd())
                f.write("cofr 0,0,0\n")  # set center of coordinates
                f.write("style stick\n")

                _inputVol = obj.getFirstItem().getVolume()
                if _inputVol is not None:
                    volID = 1
                    dim, sampling = _inputVol.getDim()[0], _inputVol.getSamplingRate()

                    f.write("open %s\n" % _inputVol.getFileName())
                    x, y, z = _inputVol.getOrigin(force=True).getShifts()
                    f.write("volume #%d style surface voxelSize %f\nvolume #%d origin "
                            "%0.2f,%0.2f,%0.2f\n"
                            % (volID, sampling, volID, x, y, z))
                else:
                    dim, sampling = 150., 1.

                bildFileName = self.protocol._getExtraPath("axis_output.bild")
                Chimera.createCoordinateAxisFile(dim, bildFileName=bildFileName, sampling=sampling)
                f.write("open %s\n" % bildFileName)

                for AS in obj:
                    f.write("open %s\n" % AS.getFileName())
                    #f.write("style stick\n")

                f.close()
                view = ChimeraView(fnCmd)
                return [view]

        elif issubclass(cls,emobj.Volume):
            view = ChimeraView(obj.getFileName())
            return [view]

        elif issubclass(cls,emobj.EMSet):
            files = obj.getFiles()
            view = ChimeraView(files)
            return [view]

        else:
            raise Exception('ChimeraViewer.visualize: '
                            'can not visualize class: %s' % obj.getClassName())


class ChimeraOldView(ChimeraView):
    """ View for calling an external command. """

    def getProgram(self):
        return emConfig.CHIMERA_OLD_BINARY_PATH

    def getEnviron(self):
        return None

class ChimeraOldViewer(ChimeraViewer):
    """ Wrapper to visualize Chimera OLD (1.10 , 1.13, ..) compatible objects . """
    _name = "Chimera"

    def __init__(self, **kwargs):
        pwviewer.Viewer.__init__(self, **kwargs)

    def _visualize(self, obj, **kwargs):
        cls = type(obj)

        if issubclass(cls, emobj.EMSet):

            files = obj.getFiles()
            view = ChimeraOldView(files)
            return [view]

        elif issubclass(cls, (emobj.Volume, emobj.AtomStruct)):
            view = ChimeraOldView(obj.getFileName())
            return [view]

        else:
            raise Exception('ChimeraOldViewer.visualize: '
                            'can not visualize class: %s' % obj.getClassName())



def mapVolsWithColorkey(displayVolFileName,
                        mapVolFileName,
                        stepColors,  # List with resolution values
                        colorList,  # List with colors
                        voldim,  # pixels
                        volOrigin=None,  # in A.
                        step=-1,  # chimera volume step (display resolution)
                        sampling=1.,
                        scriptFileName='/tmp/chimeraColor.py',
                        bgColorImage='white',  # background color
                        showAxis=True,
                        fontSize=12):  # default chimera font size
    """ colors surface of volume 'displayVolFileName' using values from
    'mapVolFikeName'. A colorkey is created using the values in stepColors
    and the color in colorList. """
    scriptFile = scriptFileName
    fhCmd = open(scriptFile, 'w')
    fhCmd.write("from chimerax.core.commands import run\n")
    fhCmd.write("run(session, 'set bgColor %s')\n" % bgColorImage)

    chimeraVolId = 1

    bildFileName = scriptFile.replace(".py", ".bild")
    Chimera.createCoordinateAxisFile(voldim[0],
                                     bildFileName=bildFileName,
                                     sampling=sampling)
    # axis
    fhCmd.write("run(session, 'open %s')\n" % bildFileName)
    if not showAxis:
        fhCmd.write("run(session, 'hide #1')\n")

    fhCmd.write("run(session, 'cofr 0,0,0')\n")  # set center of coordinates

    chimeraVolId += 1

    # first volume
    if volOrigin is None:

        x = -voldim[0] * sampling // 2
        y = -voldim[1] * sampling // 2
        z = -voldim[2] * sampling // 2
    else:
        # TODO, not sure about sign
        x = volOrigin[0]
        y = volOrigin[1]
        z = volOrigin[2]

    fhCmd.write("run(session, 'open %s')\n" % displayVolFileName)
    if step == -1:
        fhCmd.write("run(session, 'volume #%d voxelSize %s')\n" %
                    (chimeraVolId, str(sampling)))
    else:
        fhCmd.write("run(session, 'volume #%d voxelSize %s step %d')\n" %
                    (chimeraVolId, str(sampling), step))
    fhCmd.write("run(session, 'volume #%d origin %0.2f,%0.2f,%0.2f')\n"
                % (chimeraVolId, x, y, z))
    # second volume
    chimeraVolId += 1
    fhCmd.write("run(session, 'open %s')\n" % mapVolFileName)
    # TODO: Why no step here?
    fhCmd.write("run(session, 'volume #%d voxelSize %f')\n" % (chimeraVolId, sampling))
    fhCmd.write("run(session, 'volume #%d origin %0.2f,%0.2f,%0.2f')\n"
                % (chimeraVolId, x, y, z))
    fhCmd.write("run(session, 'vol #%d hide')\n" % chimeraVolId)

    # replace scolor + colorkey which has been discontinued in chimerax
    scolorStr = ''
    for step, color in zip(stepColors, colorList):
        scolorStr += '%s,%s:' % (step, color)
    scolorStr = scolorStr[:-1]
    fhCmd.write("run(session, 'color sample #%d map #%d "
                "palette " % (chimeraVolId - 1, chimeraVolId) +
                scolorStr + "')\n"
                )
    fhCmd.write(generateColorLegend(stepColors, colorList, threeLabelsOnly=False))
    fhCmd.write("run(session, 'view')\n")
    fhCmd.close()


def generateColorLegend(stepColors, colorList, threeLabelsOnly=True):
    """Return a string to write in a chimera script file a color legend - key command

    :param stepColors: list with the values for the colors
    :param colorList: list with the colors
    :param threeLabelsOnly: True, with ignore stepColors and show just 3 values: min, max and medium."""

    colorStr = 'run(session, "key{} fontSize 15 size 0.025,0.4 pos 0.01,0.3")\n'
    labelCount, keyStr = 0, ''
    stepColors.reverse(), colorList.reverse()
    if threeLabelsOnly:
        labelsIndex = [0, len(colorList) - 1, (len(colorList) - 1) // 2]

    for step, color in zip(stepColors, colorList):
      if not threeLabelsOnly or labelCount in labelsIndex:
          keyStr += ' {}:{}'.format(color, step)
      else:
          keyStr += ' {}:'.format(color)

      labelCount += 1
    return colorStr.format(keyStr)


# **************************************************************************
# *
# * Authors: Yunior C. Fonseca Reyna    (cfonseca@cnb.csic.es)
# *          Pablo Conesa Mingo         (pconesa@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import logging
logger = logging.getLogger()

from metadataviewer.dao.model import IDAO


class StarFile(IDAO):
    """
    Class to handle STAR or XMD files.
    """
    def __init__(self, inputFile):
        self._file = self.__loadFile(inputFile)
        self._tableCount = {}
        self._tableData = {}
        self._labels = {}
        self._names = []
        self._labelsTypes = {}
        self._aliases = {}

    def __loadFile(self, inputFile):
        try:
            return open(inputFile, 'r')
        except Exception as e:
            logger.error("The file could not be opened. Make sure the path is "
                         "correct: \n %s" % e)
            return None

    def fillTable(self, table, objectManager):
        """Create the table structure"""
        self._loadStarFileInfo(table)

    def fillPage(self, page, actualColumn, orderAsc):
        """
        Fill a page taking into account the page number and size.
        Variables actualColumn and orderAsc control the table order
        """
        table = page.getTable()
        tableName = table.getName()
        if table.hasSortingChanged():
            self.sort(tableName, actualColumn, orderAsc)

        # moving to the first row of the page
        pageNumber = page.getPageNumber()
        pageSize = page.getPageSize()
        firstRow = pageNumber * pageSize - pageSize
        endRow = pageNumber * pageSize  # getting more rows

        logger.debug("Creating the page with rows: %d -> %d " % (firstRow, endRow))
        for row in self._iterRowLines(tableName, firstRow, endRow):
            if row:
                page.addRow(row)

    def getTableRowCount(self, tableName):
        """Get the number of rows of a given table"""
        return self._tableCount[tableName]

    def getTableNames(self):
        """ Return all the names of the data_ blocks found in the file and
            fill all labels and data of every table name """
        if not self._names:
            f = self._file
            line = f.readline()
            logger.debug("Reading the star file completely and filling "
                         "the model().")
            while line:
                # While searching for a data line, we will store the offsets
                # for any data_ line that we find
                if line.startswith('data_'):
                    tn = line.strip().replace('data_', '')
                    self._names.append(tn)
                    data = self._getData()
                    self._tableCount[tn] = data[0]
                    self._labels[tn] = ['id'] + data[1]
                    self._labelsTypes[tn] = data[2]
                    self._tableData[tn] = data[3]
                    self._aliases[tn] = tn
                line = f.readline()

        return list(self._names)

    def getTableAliases(self):
        """Get the tables aliases"""
        return self._aliases

    def _getData(self):
        """ Method to get all information of the table (labels, data,...)"""
        self._findLabelLine()
        data = []
        line, labels = self._getLabels()
        count = 0
        f = self._file
        firstRow = line.split()
        labelsTypes = [int]
        for i in range(len(firstRow)):
            labelsTypes.append(_guessType(firstRow[i]))

        while line and not line.startswith('\n'):
            line = str(count+1) + ' ' + line
            data.append(line.split())
            count += 1
            line = f.readline().strip()
        return count, labels, labelsTypes, data

    def _getLabels(self):
        """Get the table labels"""
        logger.debug("Getting the table labels...")
        line = self._line
        labels = []
        while line.startswith('\n'):
            line = self._file.readline()
        while line.startswith('_'):
            parts = line.split()
            labels.append(parts[0][1:])
            line = self._file.readline()
        while line.startswith('\n'):
            line = self._file.readline()
        return line, labels

    def _loadStarFileInfo(self, table):
        """Create the table structure"""
        logger.debug("Creating the table columns...")
        colNames = self._labels[table.getName()]
        values = self._tableData[table.getName()][0]
        table.createColumns(colNames, values)
        table.setAlias(table.getName())

    def _findLabelLine(self):
        """Find the first labels line in the star file"""
        line = ''
        foundLoop = False

        rawLine = self._file.readline()
        while rawLine:
            if rawLine.startswith('_'):
                line = rawLine
                break
            elif rawLine.startswith('loop_'):
                foundLoop = True
            rawLine = self._file.readline()

        self._line = line.strip()
        self._foundLoop = foundLoop

    def _iterRowLines(self, tableName, firstRow, endRow):
        """Iter over the table in a range of rows """
        if self._tableCount[tableName] == 1:
            yield 1, self._tableData[tableName][0]
            return
        if firstRow + endRow > self._tableCount[tableName]:
            endRow = self._tableCount[tableName]
        for i in range(firstRow, endRow):
            values = self._tableData[tableName][i]
            yield int(values[0]), values

    def close(self):
        if getattr(self, '_file', None):
            self._file.close()
            self._file = None

    @classmethod
    def getCompatibleFileTypes(cls):
        """Return a list of compatible extension of files"""
        logger.debug("Selected StarFile DAO")
        return ['star', 'xmd']

    def sort(self, tableName, column, sortAsc=True):
        """ Sort the table in place using the provided column.
            :param column is a number, it is the index of one column. """
        _columType = self._labelsTypes[tableName][column]
        orderList = sorted(self._tableData[tableName],
                           key=lambda x: _columType(x[column]),
                           reverse=not sortAsc)
        self._tableData[tableName] = orderList

    def getSelectedRangeRowsIds(self, tableName, startRow, numberOfRows, column, reverse=True):
        """Return a range of rows starting at 'startRow' an amount of
           'numberOfRows' """
        logger.debug("Reading the table %s and selected a range of rows %d - %d" % (tableName, startRow+1, numberOfRows + 1))
        col = 0
        for i in range(len(self._labels[tableName])):
            if self._labels[tableName][i] == 'id':
                col = i
                break
        table = self._tableData[tableName]
        rowsIds = [int(table[row][col]) for row in range(startRow-1, startRow+numberOfRows)]
        return rowsIds

    def getTableWithAdditionalInfo(self):
        """Return a tuple with the table that need to show additional info and
        the column that we need to show"""
        return None

# -----------------------------------Utils methods -----------------------


def _guessType(strValue):
    try:
        int(strValue)
        return int
    except ValueError:
        try:
            float(strValue)
            return float
        except ValueError:
            return str





